import warnings
import patsy
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.tools.sm_exceptions import DomainWarning
import matplotlib.pyplot as plt

from zepid.calc.utils import probability_bounds
from zepid.causal.utils import (check_input_data, propensity_score, plot_boxplot, plot_kde, plot_love,
                                stochastic_check_conditional, standardized_mean_differences, positivity,
                                iptw_calculator)


class IPTW:
    r"""Calculates inverse probability of treatment weights. Both stabilized or unstabilized weights are implemented.
    By default, stabilized weights are stabilized by the prevalence of the treatment in the population. `IPTW` will
    also now fit the marginal structural model and estimate inverse probability of censoring weights if requested.
    Confidence intervals are calculated using robust standard errors.

    The formula for stabilized IPTW is

    .. math::

        \pi_i = \frac{\Pr(A=a)}{\Pr(A=a|L=l)}

    For unstabilized IPTW

    .. math::

        \pi_i = \frac{1}{\Pr(A=a|L=l)}

    SMR unstabilized weights for weighting to exposed (A=1)

    .. math::

        \pi_i &= 1 \;\;\text{       if}\;\; A = 1 \\
              &= \frac{\Pr(A=1|L=l)}{\Pr(A=0|L=l)} \;\;\text{if}\;\; A = 0

    For SMR weighted to the unexposed (A=0) the equation becomes

    .. math::

        \pi_i &= \frac{\Pr(A=0|L=l)}{\Pr(A=1|L=l)} \;\;\text{if}\;\; A=1 \\
              &= 1 \;\;\text{       if} \;\;A = 0

    Diagnostics are also available for generated IPTW. For a full list of diagnostics, see specific function
    documentation below. Additionally, review the references listed for an in-depth explanation

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe object containing all variables of interest
    treatment : str
        Variable name of treatment of interest. Must be coded as binary
    outcome : str
        Variable name of outcome of interest. Can be either binary or continuous
    standardize : str, optional
        Who to standardize the estimate to. Options are the entire population, the exposed, or the unexposed. See
        Sato & Matsuyama Epidemiology (2003) for details on weighting to exposed/unexposed. Weighting to the
        exposed or unexposed is also referred to as SMR weighting. Options for standardization are:
        * 'population'    :   weight to entire population
        * 'exposed'       :   weight to exposed individuals
        * 'unexposed'     :   weight to unexposed individuals
    weights: str, optional
        Optional column for weights. If specified, a weighted regression model is instead used to estimate the inverse
        probability of treatment weights. This optional is useful in the following scenario; some confounder
        information is missing and IPMW was used to correct for missing data. IPTW should be estimated with the IPMW
        to standardize to the correct pseudo-population.

    Examples
    ---------
    Setting up environment

    >>> import matplotlib.pyplot as plt
    >>> from zepid import load_sample_data, spline
    >>> from zepid.causal.ipw import IPTW
    >>> df = load_sample_data(timevary=False).drop(columns=['cd4_wk45'])
    >>> df[['cd4_rs1','cd4_rs2']] = spline(df, 'cd40', n_knots=3, term=2, restricted=True)
    >>> df[['age_rs1','age_rs2']] = spline(df, 'age0', n_knots=3, term=2, restricted=True)

    Calculate stabilized IPTW

    >>> ipt = IPTW(df, treatment='art', outcome='dead')
    >>> ipt.treatment_model('male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> ipt.marginal_structural_model('art')
    >>> ipt.fit()
    >>> ipt.summary()

    Diagnostics:

    >>> ipt.run_diagnostics()

    Calculate unstabilized IPTW weights

    >>> ipt = IPTW(df, treatment='art', outcome='dead')
    >>> ipt.treatment_model('male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', stabilized=False)
    >>> ipt.marginal_structural_model('art')
    >>> ipt.fit()
    >>> ipt.summary()

    Calculate SMR weight to the exposed population

    >>> ipt = IPTW(df, treatment='art', outcome='dead', standardize='exposed')
    >>> ipt.treatment_model('male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> ipt.marginal_structural_model('art')
    >>> ipt.fit()
    >>> ipt.summary()

    Stabilized IPTW with IPCW

    >>> ipt = IPTW(df, treatment='art', outcome='dead')
    >>> ipt.treatment_model('male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> ipt.missing_model('art + male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> ipt.marginal_structural_model('art')
    >>> ipt.fit()
    >>> ipt.summary()

    Stabilized IPTW with effect measure modifier

    >>> ipt = IPTW(df, treatment='art', outcome='dead')
    >>> ipt.treatment_model('male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', model_numerator='male')
    >>> ipt.marginal_structural_model('art + male + art:male')
    >>> ipt.fit()
    >>> ipt.summary()

    References
    ----------
    Robins JM, Hernan MA, Brumback B. (2000). Marginal structural models and causal inference in epidemiology.

    Hernán MÁ, Brumback B, Robins JM. (2000). Marginal structural models to estimate the causal effect of zidovudine
    on the survival of HIV-positive men. Epidemiology, 561-570.

    Bodnar LM, Davidian M, Siega-Riz AM, Tsiatis AA. (2004). Marginal structural models for analyzing causal effects
    of time-dependent treatments: an application in perinatal epidemiology. American Journal of
    Epidemiology, 159(10), 926-934.

    Cole SR, Hernán MA. (2008). Constructing inverse probability weights for marginal structural models.
    American journal of epidemiology, 168(6), 656-664.

    Austin PC, Stuart EA. (2015). Moving towards best practice when using inverse probability of treatment
    weighting (IPTW) using the propensity score to estimate causal treatment effects in observational studies.
    Statistics in medicine, 34(28), 3661-3679.

    Sato T, Matsuyama Y. (2003). Marginal structural models as a tool for standardization. Epidemiology, 14(6), 680-686.

    Love T. (2004). Graphical Display of Covariate Balance. Presentation,
    See http://chrp.org/love/JSM2004RoundTableHandout. pdf, 1364.
    """
    def __init__(self, df, treatment, outcome, weights=None, standardize='population'):
        self.treatment = treatment
        self.outcome = outcome
        self._missing_indicator = '__missing_indicator__'
        self.df, self._miss_flag, self._continuous_outcome = check_input_data(data=df,
                                                                              exposure=treatment,
                                                                              outcome=outcome,
                                                                              estimator="IPTW",
                                                                              drop_censoring=False,
                                                                              drop_missing=True,
                                                                              binary_exposure_only=True)
        # TODO add detection of continuous treatments

        self.average_treatment_effect = None
        self.risk_difference = None
        self.risk_ratio = None
        self.odds_ratio = None

        if standardize in ['population', 'exposed', 'unexposed']:
            self.standardize = standardize
            if standardize in ['exposed', 'unexposed']:
                warnings.warn("For the ATT and the ATU, confidence intervals calculated using the robust-variance "
                              "approach (what is currently done in zEpid) may underestimate the variance. Therefore "
                              "when requesting the ATT or the ATU, it is recommended to use bootstrapped confidence "
                              "intervals instead.", UserWarning)
        else:
            raise ValueError('Please specify one of the currently supported weighting schemes: ' +
                             'population, exposed, unexposed')

        self._weight_ = weights
        self.iptw = None
        self.ipmw = None
        self.ms_model = None
        self.__mdenom = None
        self._fit_missing_ = False
        self._miss_model = None
        self._continuous_y_type = None
        self._pos_avg = None
        self._pos_min = None
        self._pos_max = None
        self._pos_sd = None

    def treatment_model(self, model_denominator, model_numerator='1', stabilized=True, bound=False, print_results=True):
        """Logistic regression model(s) for propensity score models. The model denominator must be specified for both
        stabilized and unstabilized weights. The optional argument 'model_numerator' allows specification of the
        stabilization factor for the weight numerator. By default model results are returned

        Parameters
        ------------
        model_denominator : str
            String listing variables to predict the exposure via `patsy` syntax. For example, `'var1 + var2 + var3'`.
            This is for the predicted probabilities of the denominator
        model_numerator : str, optional
            Optional string listing variables to predict the exposure, separated by +. Only used to calculate the
            numerator. Default ('1') calculates the overall probability of exposure. In general this is recommended. If
            confounding variables are included in the numerator, they would later need to be adjusted for in the faux
            marginal structural argument. Additionally, used for assessment of effect measure modification. Argument is
            also only used when calculating stabilized weights
        stabilized : bool, optional
            Whether to return stabilized or unstabilized weights. Default is stabilized weights (True)
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            inference becomes limited to the restricted population. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation. A collection of
            floats can be provided for asymmetric trunctation
        print_results : bool, optional
            Whether to print the model results from the regression models. Default is True
        """
        # Calculating denominator probabilities
        self.__mdenom = model_denominator
        self.df['__denom__'], self.df['__numer__'], self.iptw = iptw_calculator(df=self.df,
                                                                                treatment=self.treatment,
                                                                                model_denom=model_denominator,
                                                                                model_numer=model_numerator,
                                                                                weight=self._weight_,
                                                                                stabilized=stabilized,
                                                                                standardize=self.standardize,
                                                                                bound=bound,
                                                                                print_results=print_results)

    def missing_model(self, model_denominator, model_numerator=None, stabilized=True, bound=False, print_results=True):
        """Estimation of Pr(M=0|A=a,L), which is the missing data mechanism for the outcome. The corresponding
        observation probabilities are used to account for informative censoring by observed variables. The missing_model
        only accounts for missing outcome data.

        The inverse probability weights calculated by this function account for informative censoring (missing data on
        the outcome) by observed variables. The parametric model should be sufficiently flexible to capture any
        interaction terms and functional forms of continuous variables

        Note
        ----
        The treatment variable should be included in the model

        Parameters
        ----------
        model_denominator: str
            String listing variables predicting missingness of outcomes via `patsy` syntax. For example, `
            'var1 + var2 + var3'. This is for the predicted probabilities of the denominator
        model_numerator : str, optional
            Optional string listing variables to predict the exposure, separated by +. Only used to calculate the
            numerator. Default (None) calculates the probability of censoring by treatment only. In general this is
            recommended. If assessing effect modifcation, this variable should be included in the numerator as well.
            Argument is only used when calculating stabilized weights
        stabilized : bool, optional
            Whether to use stabilized inverse probability of censoring weights
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            inference becomes limited to the restricted population. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation. A collection of
            floats can be provided for asymmetric trunctation
        print_results: bool, optional
        """
        # Error if no missing outcome data
        if not self._miss_flag:
            raise ValueError("No missing outcome data is present in the data set")

        # Warning if exposure is not included in the missingness of outcome model
        if self.treatment not in model_denominator:
            warnings.warn("For the specified missing outcome model, the exposure variable should be included in the "
                          "model", UserWarning)

        self._miss_model = self._missing_indicator + ' ~ ' + model_denominator
        fitmodel = propensity_score(self.df, self._miss_model, print_results=print_results)

        if stabilized:
            if model_numerator is None:
                mnum = self.treatment
            else:
                mnum = model_numerator
            numerator_model = propensity_score(self.df, self._missing_indicator + ' ~ ' + mnum,
                                               weights=self._weight_,
                                               print_results=print_results)
            n = numerator_model.predict(self.df)
        else:
            n = 1

        if bound:  # Bounding predicted probabilities if requested
            d = probability_bounds(fitmodel.predict(self.df), bounds=bound)
        else:
            d = fitmodel.predict(self.df)

        self.ipmw = np.where(self.df[self._missing_indicator] == 1, n / d, np.nan)
        self._fit_missing_ = True

    def marginal_structural_model(self, model):
        """Specify the marginal structural model to estimate using the inverse probability of treatment weights. The
        treatment variable should always be included in the marginal structural model. In addition to the treatment,
        effect measure modification by variables can be assessed. See the main documentation for an example of how
        to specify a effect measure modifier.

        Note
        ----
        If assessing modification, be sure to use stabilized IPTW and include the modifier in the `model_numerator`

        Parameters
        ----------
        model : str
            The specified marginal structural model to fit.
        """
        if self.treatment not in model:
            raise ValueError("The treatment variable must be specified in the marginal structural model")

        self.ms_model = model

    def fit(self, continuous_distribution='gaussian'):
        """Fit the specified marginal structural model using the calculated inverse probability of treatment weights.
        """
        if self.__mdenom is None:
            raise ValueError('No model has been fit to generated predicted probabilities')
        if self.ms_model is None:
            raise ValueError('No marginal structural model has been specified')
        if self._miss_flag and not self._fit_missing_:
            warnings.warn("All missing outcome data is assumed to be missing completely at random. To relax this "
                          "assumption to outcome data is missing at random please use the `missing_model()` "
                          "function", UserWarning)

        ind = sm.cov_struct.Independence()
        full_msm = self.outcome + ' ~ ' + self.ms_model

        df = self.df.copy()
        if self.ipmw is None:
            if self._weight_ is None:
                df['_ipfw_'] = self.iptw
            else:
                df['_ipfw_'] = self.iptw * self.df[self._weight_]
        else:
            if self._weight_ is None:
                df['_ipfw_'] = self.iptw * self.ipmw
            else:
                df['_ipfw_'] = self.iptw * self.ipmw * self.df[self._weight_]
        df = df.dropna()

        if self._continuous_outcome:
            if (continuous_distribution == 'gaussian') or (continuous_distribution == 'normal'):
                f = sm.families.family.Gaussian()
            elif continuous_distribution == 'poisson':
                f = sm.families.family.Poisson()
            else:
                raise ValueError("Only 'gaussian' and 'poisson' distributions are supported")
            self._continuous_y_type = continuous_distribution
            fm = smf.gee(full_msm, df.index, df,
                         cov_struct=ind, family=f, weights=df['_ipfw_']).fit()
            self.average_treatment_effect = pd.DataFrame()
            self.average_treatment_effect['labels'] = np.asarray(fm.params.index)
            self.average_treatment_effect.set_index(keys=['labels'], inplace=True)
            self.average_treatment_effect['ATE'] = np.asarray(fm.params)
            self.average_treatment_effect['SE(ATE)'] = np.asarray(fm.bse)
            self.average_treatment_effect['95%LCL'] = np.asarray(fm.conf_int()[0])
            self.average_treatment_effect['95%UCL'] = np.asarray(fm.conf_int()[1])

        else:
            # Ignoring DomainWarnings from statsmodels
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', DomainWarning)

                # Estimating Risk Difference
                f = sm.families.family.Binomial(sm.families.links.identity())
                fm = smf.gee(full_msm, df.index, df,
                             cov_struct=ind, family=f, weights=df['_ipfw_']).fit()
                self.risk_difference = pd.DataFrame()
                self.risk_difference['labels'] = np.asarray(fm.params.index)
                self.risk_difference.set_index(keys=['labels'], inplace=True)
                self.risk_difference['RD'] = np.asarray(fm.params)
                self.risk_difference['SE(RD)'] = np.asarray(fm.bse)
                self.risk_difference['95%LCL'] = np.asarray(fm.conf_int()[0])
                self.risk_difference['95%UCL'] = np.asarray(fm.conf_int()[1])

                # Estimating Risk Ratio
                f = sm.families.family.Binomial(sm.families.links.log())
                fm = smf.gee(full_msm, df.index, df,
                             cov_struct=ind, family=f, weights=df['_ipfw_']).fit()
                self.risk_ratio = pd.DataFrame()
                self.risk_ratio['labels'] = np.asarray(fm.params.index)
                self.risk_ratio.set_index(keys=['labels'], inplace=True)
                self.risk_ratio['RR'] = np.exp(np.asarray(fm.params))
                self.risk_ratio['SE(log(RR))'] = np.asarray(fm.bse)
                self.risk_ratio['95%LCL'] = np.exp(np.asarray(fm.conf_int()[0]))
                self.risk_ratio['95%UCL'] = np.exp(np.asarray(fm.conf_int()[1]))

                # Estimating Odds Ratio
                f = sm.families.family.Binomial()
                fm = smf.gee(full_msm, df.index, df,
                             cov_struct=ind, family=f, weights=df['_ipfw_']).fit()
                self.odds_ratio = pd.DataFrame()
                self.odds_ratio['labels'] = np.asarray(fm.params.index)
                self.odds_ratio.set_index(keys=['labels'], inplace=True)
                self.odds_ratio['OR'] = np.exp(np.asarray(fm.params))
                self.odds_ratio['SE(log(OR))'] = np.asarray(fm.bse)
                self.odds_ratio['95%LCL'] = np.exp(np.asarray(fm.conf_int()[0]))
                self.odds_ratio['95%UCL'] = np.exp(np.asarray(fm.conf_int()[1]))

    def summary(self, decimal=3):
        """Prints a summary of the results for the IPTW estimator.

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display in the result. Default is 3
        """
        print('======================================================================')
        print('              Inverse Probability of Treatment Weights                ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.treatment, self.df.shape[0]))

        fmt = 'Outcome:          {:<15} No. Missing Outcome:  {:<20}'
        print(fmt.format(self.outcome, np.sum(self.df[self.outcome].isnull())))

        fmt = 'g-Model:          {:<15} Missing Model:        {:<20}'
        if self._fit_missing_:
            m = 'Logistic'
        else:
            m = 'None'
        print(fmt.format('Logistic', m))
        print('======================================================================')

        if self._continuous_outcome:
            print('Average Treatment Effect')
            print('----------------------------------------------------------------------')
            print(np.round(self.average_treatment_effect, decimals=decimal))
        else:
            print('Risk Difference')
            print('----------------------------------------------------------------------')
            print(np.round(self.risk_difference, decimals=decimal))
            print('----------------------------------------------------------------------')
            print('Risk Ratio')
            print(np.round(self.risk_ratio, decimals=decimal))
            print('----------------------------------------------------------------------')
            print('Odds Ratio')
            print(np.round(self.odds_ratio, decimals=decimal))
        print('======================================================================')

    def run_diagnostics(self, iptw_only=True):
        """Run all currently implemented diagnostics for inverse probability of treatment weights available. Each
        diagnostic can be called individually for further optional specifications. `run_weight_diagnostics` will
        provide results for all implemented diagnostics for ease of the user. For presentation of results, I recommend
        calling each function individually and utilizing the optional parameters

        Note
        ----
        The plot presented cannot be edited. To edit the plots, call `plot_kde` or `plot_love` directly. Those
        functions return an axes object

        Parameters
        ----------
        iptw_only : bool, optional
            Whether to include weights generated from missing_model in the diagnostics. Default is True, which runs the
            diagnostics for IPTW only

        Returns
        -------
        None
        """
        self.positivity(iptw_only=iptw_only)

        print('\n======================================================================')
        print('                Standardized Mean Differences')
        print('======================================================================')
        print(self.standardized_mean_differences(iptw_only=iptw_only).set_index(keys='labels'))
        print('======================================================================')

        plt.figure(figsize=[9, 4])
        plt.subplot(122)
        self.plot_kde()
        plt.title("Kernel Density of Propensity Scores")

        plt.subplot(121)
        self.plot_love(iptw_only=iptw_only)
        plt.title("Love Plot")
        plt.tight_layout()
        plt.show()

    def plot_kde(self, measure='probability', bw_method='scott', fill=True, color_e='b', color_u='r'):
        """Generates a density plot that can be used to check whether positivity may be violated qualitatively. The
        kernel density used is SciPy's Gaussian kernel. Either Scott's Rule or Silverman's Rule can be implemented.
        Alternative option to the boxplot of probabilities

        Parameters
        ------------
        measure : str, optional
            Measure to plot. Options include either the probabilities or log-odds stratified by treatment received.
            Default is probabilities (measure='probability'). Log-odds can be requested via measure='logit'
        bw_method : str, optional
            Method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
        fill : bool, optional
            Whether to color the area under the density curves. Default is true
        color_e : str, optional
            Color of the line/area for the treated group. Default is Blue
        color_u : str, optional
            Color of the line/area for the treated group. Default is Red

        Returns
        ---------------
        matplotlib axes
        """
        ax = plot_kde(df=self.df, treatment=self.treatment, probability='__denom__', measure=measure,
                      bw_method=bw_method, fill=fill, color_e=color_e, color_u=color_u)
        return ax

    def plot_boxplot(self, measure='probability'):
        """Generates a stratified boxplot that can be used to visually check whether positivity may be violated,
        qualitatively. Alternative option to the kernel density plot.

        Parameters
        ----------
        measure : str, optional
            Measure to plot. Options include either the probabilities or log-odds stratified by treatment received.
            Default is probabilities (measure='probability'). Log-odds can be requested via measure='logit'

        Returns
        -------------
        matplotlib axes
        """
        ax = plot_boxplot(df=self.df, treatment=self.treatment, probability='__denom__', measure=measure)
        return ax

    def positivity(self, decimal=3, iptw_only=True):
        """Use this to assess whether positivity is a valid assumption. For stabilized weights, the mean weight should
        be approximately 1. For unstabilized weights, the mean weight should be approximately 2. If there are extreme
        outliers, this may indicate problems with the calculated weights

        Parameters
        --------------
        decimal : int, optional
            Number of decimal places to display. Default is three
        iptw_only : bool, optional
            Whether the diagnostic should be run on IPTW only or include weights from the missing_model. Default is
            True, which returns IPTW only

        Returns
        --------------
        None
        """
        if iptw_only:
            df = self.df
            df['_ipfw_'] = self.iptw
        else:
            df = self.df
            df['_ipfw_'] = self.iptw * self.ipmw

        self._pos_avg, self._pos_sd, self._pos_min, self._pos_max = positivity(df=df, weights='_ipfw_')
        print('======================================================================')
        print('                     Weight Positivity Diagnostics')
        print('======================================================================')
        print('If the mean of the weights is far from either the min or max, this may\n '
              'indicate the model is incorrect or positivity is violated')
        print('Average weight should be')
        print('\t1.0 for stabilized')
        print('\t2.0 for unstabilized')
        print('----------------------------------------------------------------------')
        print('Mean weight:           ', round(self._pos_avg, decimal))
        print('Standard Deviation:    ', round(self._pos_sd, decimal))
        print('Minimum weight:        ', round(self._pos_min, decimal))
        print('Maximum weight:        ', round(self._pos_max, decimal))
        print('======================================================================')

    def standardized_mean_differences(self, iptw_only=True):
        """Calculates the standardized mean differences for all variables. Default calculates the standardized mean
        difference for all variables included in the IPTW denominator

        Parameters
        ----------
        iptw_only : bool, optional
            Whether the diagnostic should be run on IPTW only or include weights from the missing_model. Default is
            True, which returns IPTW only

        Returns
        -------
        DataFrame
            Returns pandas DataFrame of calculated standardized mean differences. Columns are labels (variables labels),
            smd_u (unweighted standardized difference), and smd_w (weighted standardized difference)
        """
        if iptw_only:
            df = self.df
            df['_ipfw_'] = self.iptw
        else:
            df = self.df
            df['_ipfw_'] = self.iptw * self.ipmw

        s = standardized_mean_differences(df=df, treatment=self.treatment, weight='_ipfw_', formula=self.__mdenom)
        return s

    def plot_love(self, color_unweighted='r', color_weighted='b', shape_unweighted='o', shape_weighted='o',
                  iptw_only=True):
        """Generates a Love-plot to detail covariate balance based on the IPTW weights. Further details on the usage of
        this plot are available in Austin PC & Stuart EA 2015 https://onlinelibrary.wiley.com/doi/full/10.1002/sim.6607

        The Love plot generates a dashed line at standardized mean difference of 0.10. Ideally, weighted SMD are below
        this level. Below 0.20 may also be sufficient. Variables above this level may be unbalanced despite the
        weighting procedure. Different functional forms (or approaches like machine learning) may be worth considering

        Parameters
        ----------
        color_unweighted : str, optional
            Color for the unweighted standardized mean differences. Default is red
        color_weighted : str, optional
            Color for the weighted standardized mean differences. Default is blue
        shape_unweighted : str, optional
            Shape of points for the unweighted standardized mean differences. Default is circles
        shape_weighted:
            Shape of points for the weighted standardized mean differences. Default is circles
        iptw_only : bool, optional
            Whether the diagnostic should be run on IPTW only or the weights multiplied together. Default is IPTW only

        Returns
        -------
        matplotlib axes
        """
        if iptw_only:
            df = self.df
            df['_ipfw_'] = self.iptw
        else:
            df = self.df
            df['_ipfw_'] = self.iptw * self.ipmw

        ax = plot_love(df=self.df, treatment=self.treatment, weight='_ipfw_', formula=self.__mdenom,
                       color_unweighted=color_unweighted, color_weighted=color_weighted,
                       shape_unweighted=shape_unweighted, shape_weighted=shape_weighted)
        return ax


class StochasticIPTW:
    r"""Calculates the IPTW estimate for stochastic treatment plans. `StochasticIPTW` will returns the estimated
    marginal outcome for that treatment plan. This is distinct from `IPTW`, which returns an array of weights. For
    confidence intervals, a bootstrapping procedure needs to be used.

    The formula for IPTW for a stochastic treatment is

    .. math::

        \pi_i = \frac{\overline{\Pr}(A=a|L)}{\Pr(A=a|L)}

    where :math:`\bar{\Pr}` is the new probability of treatment under the proposed stochastic treatment. This
    probability can be unconditional (everyone treated at some constant percent) or it can be conditional on observed
    covariates. The denominator is the same estimated probability of treatment in the standard IPTW formula. Basically,
    we are manipulating how many the treated individuals represent in a new pseudo-population

    Note
    ----
    `StochasticIPTW` estimates the marginal outcome at a specified treatment distribution. Unlike IPTW, it does not
    immediately result in a comparison between two treatment levels (i.e. we are not estimating a marginal structural
    model in this case). For a comparison, two different version would need to be specified.

    `StochasticIPTW` does not contain the diagnostics that are contained within `IPTW`. This IPTW estimation approach
    makes weaker assumptions regarding positivity and causal consistency.

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe object containing all variables of interest
    treatment : str
        Variable name of treatment variable of interest. Must be coded as binary. 1 should indicate treatment,
        while 0 indicates no treatment
    outcome: str
        Variable name of outcome variable of interest. May be binary or continuous.
    weights: str, optional
        Optional column for weights. If specified, a weighted regression model is instead used to estimate the inverse
        probability of treatment weights. This optional is useful in the following scenario; some confounder
        information is missing and IPMW was used to correct for missing data. IPTW should be estimated with the IPMW
        to standardize to the correct pseudo-population.

    Examples
    --------
    Loading data

    >>> from zepid import load_sample_data, spline
    >>> from zepid.causal.ipw import StochasticIPTW
    >>> df = load_sample_data(timevary=False).drop(columns=['cd4_wk45'])
    >>> df[['cd4_rs1','cd4_rs2']] = spline(df,'cd40',n_knots=3,term=2,restricted=True)
    >>> df[['age_rs1','age_rs2']] = spline(df,'age0',n_knots=3,term=2,restricted=True)

    Estimating marginal outcome under treatment plan where 80% are randomly treated

    >>> ipw = StochasticIPTW(df, treatment='art', outcome='dead')
    >>> ipw.treatment_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> ipw.fit(p=0.8)
    >>> ipw.summary()

    Estimating marginal outcome under treatment plan where 10% are randomly treated

    >>> ipw.fit(p=0.1)
    >>> ipw.summary()

    Estimating marginal outcome under treatment plan where 75% of men are treated and 90% of women

    >>> ipw.fit(p=[0.75, 0.90], conditional=["df['male']==1", "df['male']==0"])
    >>> ipw.summary()

    References
    ----------
    Muñoz ID & van der Laan M. (2012). Population intervention causal effects based on stochastic interventions.
    Biometrics, 68(2), 541-549.
    """
    def __init__(self, df, treatment, outcome, weights=None):
        self.treatment = treatment
        self.outcome = outcome
        self._missing_indicator = '__missing_indicator__'
        self.df, self._miss_flag, self._continuous_outcome = check_input_data(data=df,
                                                                              exposure=treatment,
                                                                              outcome=outcome,
                                                                              estimator="StochasticIPTW",
                                                                              drop_censoring=True,
                                                                              drop_missing=True,
                                                                              binary_exposure_only=True)
        self.weights = weights
        self.marginal_outcome = np.nan
        self._pdenom_ = None

    def treatment_model(self, model, print_results=True):
        r"""Specify the parametric regression model for the observed treatment conditional on the sufficient adjustment
        set. This model estimates the following component of the stochastic IPTW weights

        .. math::

            \widehat{\Pr}(A=a|L)

        Parameters
        ----------
        model : str
            String listing variables to predict the exposure via `patsy` syntax. For example, `'var1 + var2 + var3'`
        print_results : bool, optional
            Whether to print the model results from the regression models. Default is True
        """
        # Calculating denominator probabilities
        denominator_model = propensity_score(self.df, self.treatment + ' ~ ' + model,
                                             weights=self.weights,
                                             print_results=print_results)
        self._pdenom_ = denominator_model.predict(self.df)

    def fit(self, p, conditional=None):
        """Estimates the mean outcome under the specified stochastic treatment plan.As currently implemented, `p`
        percent of the population is randomly treated. Unlike the stochastic g-formula, we only need to estimate the
        marginal outcome once.

        Parameters
        ----------
        p : float, list
            Percent of the population to randomly treat
        conditional : list, optional
            Exclusive conditions to place on the data set for treatment percents. If specified, must match the length
            of the list of probabilities in `p`. For specification of conditions, the data set should be referred to
            as `df` in strings. For example, `"df['male']==1"` restricts that probability to only males

        Returns
        -------
        marginal_outcome
            Gains marginal outcome attribute. Summary function can also be called afterwards
        """
        p = np.array(p)

        if self._pdenom_ is None:
            raise ValueError("The treatment_model() function must be specified before the fit() function")
        if np.any(p > 1) or np.any(p < 0):
            raise ValueError("All specified treatment probabilities must be between 0 and 1")
        if conditional is not None:
            if len(p) != len(conditional):
                raise ValueError("'p' and 'conditional' must be the same length")

        df = self.df.copy()

        if conditional is None:
            df['_numer_'] = np.where(df[self.treatment] == 1, p, 1 - p)
        else:
            stochastic_check_conditional(df=df, conditional=conditional)
            df['_numer_'] = np.nan
            for c, prop in zip(conditional, p):
                df['_numer_'] = np.where(eval(c),
                                         np.where(df[self.treatment] == 1, prop, 1 - prop),
                                         df['_numer_'])

        df['_denom_'] = np.where(df[self.treatment] == 1, self._pdenom_, 1 - self._pdenom_)
        df['_ipw_'] = df['_numer_'] / df['_denom_']

        if self.weights is not None:
            df['_ipw_'] = df['_ipw_'] * df[self.weights]

        self.marginal_outcome = np.average(df[self.outcome], weights=df['_ipw_'])

    def summary(self, decimal=3):
        """Prints the summary information for the marginal outcome under the treatment plan of interest.

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if np.isnan(self.marginal_outcome):
            raise ValueError('The fit() function must be specified before summary()')

        print('======================================================================')
        print('                       Stochastic IPTW                                ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.treatment, self.df.shape[0]))
        fmt = 'Outcome:          {:<15} Treatment Model:      {:<20}'
        print(fmt.format(self.outcome, 'Logistic'))
        print('======================================================================')
        print('Risk:  ', round(self.marginal_outcome, decimal))
        print('======================================================================')



import copy
import warnings
import patsy
import numpy as np
import pandas as pd
from scipy.stats import logistic, norm
import statsmodels.api as sm
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from numpy.random import RandomState

from zepid.calc.utils import probability_bounds, probability_to_odds, odds_to_probability
from zepid.causal.utils import check_input_data, aipw_calculator
from zepid.causal.doublyrobust.utils import tmle_unit_unbound, tmle_unit_bounds


class SingleCrossfitAIPTW:
    """Implementation of the Augmented Inverse Probability Weighting estimator with a cross-fit procedure. The purpose
    of the cross-fit procedure is to all for non-Donsker nuisance function estimators. Some of machine learning
    algorithms are non-Donsker. In practice this means that confidence interval coverage can be incorrect when certain
    nuisance function estimators are used. Additionally, bias may persist as well. Cross-fitting is meant to alleviate
    this issue, therefore cross-fitting with a doubly-robust estimator is recommended when using machine learning.

    `SingleCrossfitAIPTW` uses a single cross-fit procedure, where the data set is paritioned into at least two
    non-overlapping splits. The nuisance function estimators are then estimated in each split. The estimated nuisance
    functions are then used to predict values in a non-overlapping split. This decouple the nuisance function estimation
    from the data used to estimate it

    Note
    ----
    Because of the repetitions of the procedure are needed to reduce variance determined by a particular partition, it
    can take a long time to run this code.

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing all necessary variables
    exposure : str
        Label for treatment column in the pandas data frame
    outcome : str
        Label for outcome column in the pandas data frame
    alpha : float, optional
        Alpha for confidence interval level. Default is 0.05

    Examples
    --------
    Setting up environment

    >>> from sklearn.linear_model import LogisticRegression
    >>> from zepid import load_sample_data
    >>> from zepid.causal.doublyrobust import SingleCrossfitAIPTW
    >>> df = load_sample_data(False).drop(columns='cd4_wk45').dropna()

    Estimating the single cross-fit AIPTW

    >>> scaipw = SingleCrossfitAIPTW(df, exposure='art', outcome='dead')
    >>> scaipw.exposure_model("male + age0 + cd40 + dvl0", estimator=LogisticRegression(solver='lbfgs'))
    >>> scaipw.outcome_model("art + male + age0 + cd40 + dvl0", estimator=LogisticRegression(solver='lbfgs'))
    >>> scaipw.fit(n_splits=5, n_partitions=100)
    >>> scaipw.summary()

    References
    ----------
    Chernozhukov V, Chetverikov D, Demirer M, Duflo E, Hansen C, Newey W, & Robins J. (2018). "Double/debiased machine
    learning for treatment and structural parameters". The Econometrics Journal 21:1; pC1–C6
    """
    def __init__(self, df, exposure, outcome, alpha=0.05):
        self.exposure = exposure
        self.outcome = outcome
        self.df, self._miss_flag, self._continuous_outcome_ = check_input_data(data=df,
                                                                               exposure=exposure,
                                                                               outcome=outcome,
                                                                               estimator="SingleCrossfitAIPTW",
                                                                               drop_censoring=True,
                                                                               drop_missing=True,
                                                                               binary_exposure_only=True)
        self.alpha = alpha

        self._a_covariates = None
        self._y_covariates = None
        self._a_estimator = None
        self._y_estimator = None
        self._fit_treatment_ = False
        self._fit_outcome_ = False
        self._gbounds = None
        self._n_splits_ = 0
        self._n_partitions = 0
        self._combine_method_ = None

        self.ace_vector = None
        self.ace_var_vector = None
        self.ace = None
        self.ace_ci = None
        self.ace_se = None

        self.risk_difference_vector = None
        self.risk_difference_var_vector = None
        self.risk_difference = None
        self.risk_difference_ci = None
        self.risk_difference_se = None

        self.risk_ratio_vector = None
        self.risk_ratio_var_vector = None
        self.risk_ratio = None
        self.risk_ratio_ci = None
        self.risk_ratio_se = None

    def exposure_model(self, covariates, estimator, bound=False):
        """Specify the treatment nuisance model variables and estimator(s) to use. These parameters are held
        in the background until the .fit() function is called. These approaches are for used each sample split

        Parameters
        ----------
        covariates : str
            Confounders to include in the propensity score model. Follows patsy notation
        estimator :
            Estimator to use for prediction of the propensity score
        bound : float, list, optional
            Whether to bound predicted probabilities. Default is False, which does not bound
        """
        self._a_estimator = estimator
        self._a_covariates = covariates
        self._fit_treatment_ = True
        self._gbounds = bound

    def outcome_model(self, covariates, estimator):
        """Specify the outcome nuisance model variables and estimator(s) to use. These parameters are held
        in the background until the .fit() function is called. These approaches are for used each sample split

        Parameters
        ----------
        covariates : str
            Confounders to include in the propensity score model. Follows patsy notation
        estimator :
            Estimator to use for prediction of the propensity score
        """
        self._y_estimator = estimator
        self._y_covariates = covariates
        self._fit_outcome_ = True

    def fit(self, n_splits=2, n_partitions=100, method='median', random_state=None):
        """Runs the crossfit estimation procedure with augmented inverse probability weighting estimator. The
        estimation process is completed for multiple different splits during the procedure. The final estimate is
        defined as either the median or mean of the causal measure from each of the different splits. Median is
        used as the default since it is more stable.

        Note
        ----
        `n_partition` should be kept high to reduce dependency of results on the chosen number of splits

        Confidence intervals come from influences curves and incorporates the within-split variance and between-split
        variance.

        Parameters
        ----------
        n_splits : int
            Number of splits to use with a default of 2. The number of splits must be greater than or equal to 2.
        n_partitions : int
            Number of times to repeat the partition process. The default is 100, which I have seen good performance
            with in the past. Note that this algorithm can take a long time to run for high values of this parameter.
            It is best to test out run-times on small numbers first. Also if running in parallel, it can be reduced
        method : str, optional
            Method to obtain point estimates and standard errors. Median method takes the median (which is more robust)
            and the mean takes the mean. It has been remarked that the median is preferred, since it is more stable to
            extreme outliers, which may happen in finite samples
        random_state : None, int, optional
            Whether to set a seed for the partitions. Default is None (which does not use a user-set seed). Any valid
            NumPy seed can be input. Note that you should also state the random_state of all (applicable) estimators
            to ensure replicability. Seeds are chosen by the following procedure. The input random_state is based to
            np.random.choice to select n_partitions between 0 and 5million. That list of n_partition-length is then
            passed to each iteration of the cross-fitting pandas.DataFrame.sample(random_state).
        """
        # Checking for various issues
        if not self._fit_treatment_:
            raise ValueError("exposure_model() must be called before fit()")
        if not self._fit_outcome_:
            raise ValueError("outcome_model() must be called before fit()")
        if n_splits < 2:
            raise ValueError("SingleCrossfitAIPTW requires that n_splits > 1")

        # Storing some information
        self._n_splits_ = n_splits
        self._n_partitions = n_partitions
        self._combine_method_ = method

        # Creating blank lists
        diff_est, diff_var, ratio_est, ratio_var = [], [], [], []

        # Conducts the re-sampling procedure
        if random_state is None:
            random_state = [None] * n_partitions
        else:
            random_state = RandomState(random_state).choice(range(5000000), size=n_partitions, replace=False)

        for j in range(self._n_partitions):
            # Estimating for a particular split (lots of functions happening in the background)
            result = self._single_crossfit_(random_state=random_state[j])

            # Appending results of this particular split combination
            diff_est.append(result[0])
            diff_var.append(result[1])
            if not self._continuous_outcome_:
                ratio_est.append(result[2])
                ratio_var.append(result[3])

        # Obtaining overall estimate and (1-alpha)% CL from all splits
        zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        est, var = calculate_joint_estimate(diff_est, diff_var, method=method)
        if self._continuous_outcome_:
            self.ace_vector = diff_est
            self.ace_var_vector = diff_var
            self.ace = est
            self.ace_se = np.sqrt(var)
            self.ace_ci = (self.ace - zalpha*self.ace_se,
                           self.ace + zalpha*self.ace_se)
        else:
            # Risk Difference
            self.risk_difference_vector = diff_est
            self.risk_difference_var_vector = diff_var
            self.risk_difference = est
            self.risk_difference_se = np.sqrt(var)
            self.risk_difference_ci = (self.risk_difference - zalpha*self.risk_difference_se,
                                       self.risk_difference + zalpha*self.risk_difference_se)
            # Risk Ratio
            self.risk_ratio_vector = ratio_est
            self.risk_ratio_var_vector = ratio_var
            ln_rr, ln_rr_var = calculate_joint_estimate(np.log(self.risk_ratio_vector),
                                                        self.risk_ratio_var_vector, method=method)
            self.risk_ratio = np.exp(ln_rr)
            self.risk_ratio_se = np.sqrt(ln_rr_var)
            self.risk_ratio_ci = (np.exp(ln_rr - zalpha*self.risk_ratio_se),
                                  np.exp(ln_rr + zalpha*self.risk_ratio_se))

    def summary(self, decimal=3):
        """Prints summary of model results

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if (self._fit_outcome_ is False) or (self._fit_treatment_ is False):
            raise ValueError('exposure_model and outcome_model must be specified before the estimate can '
                             'be generated')

        print('======================================================================')
        print('                     Single Cross-fit AIPTW                           ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.exposure, self.df.shape[0]))
        fmt = 'Outcome:          {:<15} No. of Splits:        {:<20}'
        print(fmt.format(self.outcome, self._n_splits_))
        fmt = 'Method:           {:<15} No. of Partitions:    {:<20}'
        print(fmt.format(self._combine_method_, self._n_partitions))

        print('======================================================================')
        if self._continuous_outcome_:
            print('Average Causal Effect: ', round(float(self.ace), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.ace_ci[0], decimal)), ',',
                  str(round(self.ace_ci[1], decimal)) + ')')
        else:
            print('Risk Difference:    ', round(float(self.risk_difference), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_difference_ci[0], decimal)), ',',
                  str(round(self.risk_difference_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Risk Ratio:         ', round(float(self.risk_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_ratio_ci[0], decimal)), ',',
                  str(round(self.risk_ratio_ci[1], decimal)) + ')')

        print('======================================================================')

    def run_diagnostics(self, color='gray'):
        """Runs available diagnostics for the plots. Currently diagnostics consist of a plot of the different point
        estimates and variance estimates across different partitions. Diagnostics for cross-fit estimators is ongoing.
        If you have any suggestions, please feel free to contact me on GitHub

        Parameters
        ----------
        color : str, optional
            Controls color of the plots. Default is gray

        Returns
        -------
        Plot to console
        """
        # Continuous outcomes have less plots to generate
        if self._continuous_outcome_:
            _run_diagnostic_(diff=self.ace_vector, diff_var=self.ace_var_vector,
                             color=color)

        # Binary outcomes have plots for all measures
        else:
            _run_diagnostic_(diff=self.risk_difference_vector, diff_var=self.risk_difference_var_vector,
                             rratio=self.risk_ratio_vector, rratio_var=self.risk_ratio_var_vector,
                             color=color)

    def _single_crossfit_(self, random_state):
        """Background function that runs a single crossfit of the split samples
        """
        # Dividing into s different splits
        sample_split = _sample_split_(self.df, n_splits=self._n_splits_, random_state=random_state)

        # Determining pairings to use for each sample split and each combination
        pairing_exposure = [i - 1 for i in range(self._n_splits_)]
        pairing_outcome = pairing_exposure

        # Estimating treatment nuisance model
        a_models = _treatment_nuisance_(treatment=self.exposure, estimator=self._a_estimator,
                                        samples=sample_split, covariates=self._a_covariates)
        # Estimating outcome nuisance model
        y_models = _outcome_nuisance_(outcome=self.outcome, estimator=self._y_estimator,
                                      samples=sample_split, covariates=self._y_covariates)

        # Generating predictions based on set pairs for cross-fit procedure
        predictions = []
        y_obs, a_obs = np.array([]), np.array([])
        split_index = []
        for id, ep, op in zip(range(self._n_splits_), pairing_exposure, pairing_outcome):
            predictions.append(self._generate_predictions_(sample_split[id],
                                                           a_model_v=a_models[ep],
                                                           y_model_v=y_models[op]))
            # Generating vector of Y in correct order
            y_obs = np.append(y_obs, np.asarray(sample_split[id][self.outcome]))
            # Generating vector of A in correct order
            a_obs = np.append(a_obs, np.asarray(sample_split[id][self.exposure]))
            # Generating index for splits
            split_index.extend([id]*sample_split[id].shape[0])

        # Stacking Predicted Pr(A=1), Y(a=1), Y(a=0)
        pred_a_array, pred_y1_array, pred_y0_array = np.array([]), np.array([]), np.array([])
        for preds in predictions:
            pred_a_array = np.append(pred_a_array, preds[0])
            pred_y1_array = np.append(pred_y1_array, preds[1])
            pred_y0_array = np.append(pred_y0_array, preds[2])

        # Applying bounds if requested
        if self._gbounds:  # Bounding g-model if requested
            pred_a_array = probability_bounds(pred_a_array, bounds=self._gbounds)

        # Calculating point estimates
        difference, var_diff = aipw_calculator(y=y_obs, a=a_obs,
                                               py_a=pred_y1_array, py_n=pred_y0_array,
                                               pa1=pred_a_array, pa0=1-pred_a_array,
                                               splits=np.asarray(split_index),
                                               difference=True, continuous=self._continuous_outcome_)
        if self._continuous_outcome_:
            return difference, var_diff
        else:
            ratio, var_ratio = aipw_calculator(y=y_obs, a=a_obs,
                                               py_a=pred_y1_array, py_n=pred_y0_array,
                                               pa1=pred_a_array, pa0=1 - pred_a_array,
                                               splits=np.asarray(split_index),
                                               difference=False, continuous=False)
            return difference, var_diff, ratio, var_ratio

    def _generate_predictions_(self, sample, a_model_v, y_model_v):
        """Generates predictions from fitted functions (in background of _single_crossfit()
        """
        s = sample.copy()

        # Predicting Pr(A=1|L)
        xdata = np.asarray(patsy.dmatrix(self._a_covariates + ' - 1', s))
        a_pred = _ml_predictor(xdata, fitted_algorithm=a_model_v)

        # Predicting E(Y|A=1, L)
        s[self.exposure] = 1
        xdata = np.asarray(patsy.dmatrix(self._y_covariates + ' - 1', s))
        y_treat = _ml_predictor(xdata, fitted_algorithm=y_model_v)

        # Predicting E(Y|A=0, L)
        s[self.exposure] = 0
        xdata = np.asarray(patsy.dmatrix(self._y_covariates + ' - 1', s))
        y_none = _ml_predictor(xdata, fitted_algorithm=y_model_v)

        return a_pred, y_treat, y_none


class DoubleCrossfitAIPTW:
    """Implementation of the augmented inverse probability weighted estimator with a double cross-fit procedure. The
    purpose of the cross-fit procedure is to all for non-Donsker nuisance function estimators. Some of machine learning
    algorithms are non-Donsker. In practice this means that confidence interval coverage can be incorrect when certain
    nuisance function estimators are used. Additionally, bias may persist as well. Cross-fitting is meant to alleviate
    this issue, therefore cross-fitting with a doubly-robust estimator is recommended when using machine learning.

    `DoubleCrossfitAIPTW` allows for double cross-fitting, where the data set is partitioned into at least three
    non-overlapping splits. The nuisance function estimators are then estimated in each split. The estimated nuisance
    functions are then used to predict values in the opposing split. Different splits are used for each nuisance
    function. A double cross-fit procedure further de-couples the nuisance function estimation compared to single
    cross-fit procedures.

    Note
    ----
    Because of the repetitions of the procedure are needed to reduce variance determined by a particular partition, it
    can take a long time to run this code. On a data set of 3000 observations with 100 different partitions it takes
    about an hour. The advantage is that the code can be ran in parallel. See the documentation for an example.

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing all necessary variables
    exposure : str
        Label for treatment column in the pandas data frame
    outcome : str
        Label for outcome column in the pandas data frame
    alpha : float, optional
        Alpha for confidence interval level. Default is 0.05

    Examples
    --------
    Setting up environment

    >>> from sklearn.linear_model import LogisticRegression
    >>> from zepid import load_sample_data
    >>> from zepid.causal.doublyrobust import SingleCrossfitAIPTW
    >>> df = load_sample_data(False).drop(columns='cd4_wk45').dropna()

    Estimating the double cross-fit AIPTW

    >>> dcaipw = DoubleCrossfitAIPTW(df, exposure='art', outcome='dead')
    >>> dcaipw.exposure_model("male + age0 + cd40 + dvl0", estimator=LogisticRegression(solver='lbfgs'))
    >>> dcaipw.outcome_model("art + male + age0 + cd40 + dvl0", estimator=LogisticRegression(solver='lbfgs'))
    >>> dcaipw.fit(n_splits=5, n_partitions=100)
    >>> dcaipw.summary()

    References
    ----------
    Newey WK, Robins JR. (2018) "Cross-fitting and fast remainder rates for semiparametric estimation".
    arXiv:1801.09138

    Zivich PN, & Breskin A. (2020). Machine learning for causal inference: on the use of cross-fit estimators.
    arXiv preprint arXiv:2004.10337.

    Chernozhukov V, Chetverikov D, Demirer M, Duflo E, Hansen C, Newey W, & Robins J. (2018). "Double/debiased machine
    learning for treatment and structural parameters". The Econometrics Journal 21:1; pC1–C6
    """
    def __init__(self, df, exposure, outcome, alpha=0.05):
        self.exposure = exposure
        self.outcome = outcome
        self.df, self._miss_flag, self._continuous_outcome_ = check_input_data(data=df,
                                                                               exposure=exposure,
                                                                               outcome=outcome,
                                                                               estimator="DoubleCrossfitAIPTW",
                                                                               drop_censoring=True,
                                                                               drop_missing=True,
                                                                               binary_exposure_only=True)
        self.alpha = alpha

        self._a_covariates = None
        self._y_covariates = None
        self._a_estimator = None
        self._y_estimator = None
        self._fit_treatment_ = False
        self._fit_outcome_ = False
        self._gbounds = None
        self._n_splits_ = 0
        self._n_partitions = 0
        self._combine_method_ = None

        self.ace_vector = None
        self.ace_var_vector = None
        self.ace = None
        self.ace_ci = None
        self.ace_se = None

        self.risk_difference_vector = None
        self.risk_difference_var_vector = None
        self.risk_difference = None
        self.risk_difference_ci = None
        self.risk_difference_se = None

        self.risk_ratio_vector = None
        self.risk_ratio_var_vector = None
        self.risk_ratio = None
        self.risk_ratio_ci = None
        self.risk_ratio_se = None

    def exposure_model(self, covariates, estimator, bound=False):
        """Specify the treatment nuisance model variables and estimator(s) to use. These parameters are held
        in the background until the .fit() function is called. These approaches are for used each sample split

        Parameters
        ----------
        covariates : str
            Confounders to include in the propensity score model. Follows patsy notation
        estimator :
            Estimator to use for prediction of the propensity score
        bound : float, list, optional
            Whether to bound predicted probabilities. Default is False, which does not bound
        """
        self._a_estimator = estimator
        self._a_covariates = covariates
        self._fit_treatment_ = True
        self._gbounds = bound

    def outcome_model(self, covariates, estimator):
        """Specify the outcome nuisance model variables and estimator(s) to use. These parameters are held
        in the background until the .fit() function is called. These approaches are for used each sample split

        Parameters
        ----------
        covariates : str
            Confounders to include in the propensity score model. Follows patsy notation
        estimator :
            Estimator to use for prediction of the propensity score
        """
        self._y_estimator = estimator
        self._y_covariates = covariates
        self._fit_outcome_ = True

    def fit(self, n_splits=3, n_partitions=100, method='median', random_state=None):
        """Runs the crossfit estimation procedure with augmented inverse probability weighted estimator. The
        estimation process is completed for multiple different splits during the procedure. The final estimate is
        defined as either the median or mean of the average causal effect from each of the different splits. Median is
        used as the default since it is more stable.

        Note
        ----
        `n_partition` should be kept high to reduce dependency of results on the chosen number of splits

        Confidence intervals come from influences curves and incorporates the within-split variance and between-split
        variance.

        Parameters
        ----------
        n_splits : int
            Number of splits to use. The default is 3, which is valid for both single cross-fit and double cross-fit.
            Single cross-fit is also compatible with 2 as the the number of splits
        n_partitions : int
            Number of times to repeat the partition process. The default is 100, which I have seen good performance
            with in the past. Note that this algorithm can take a long time to run for high values of this parameter.
            It is best to test out run-times on small numbers first. Also if running in parallel, it can be reduced
        method : str, optional
            Method to obtain point estimates and standard errors. Median method takes the median (which is more robust)
            and the mean takes the mean. It has been remarked that the median is preferred, since it is more stable to
            extreme outliers, which may happen in finite samples
        random_state : None, int, optional
            Whether to set a seed for the partitions. Default is None (which does not use a user-set seed). Any valid
            NumPy seed can be input. Note that you should also state the random_state of all (applicable) estimators
            to ensure replicability. Seeds are chosen by the following procedure. The input random_state is based to
            np.random.choice to select n_partitions between 0 and 5million. That list of n_partition-length is then
            passed to each iteration of the cross-fitting pandas.DataFrame.sample(random_state).
        """
        # Checking for various issues
        if not self._fit_treatment_:
            raise ValueError("exposure_model() must be called before fit()")
        if not self._fit_outcome_:
            raise ValueError("outcome_model() must be called before fit()")
        if n_splits < 3:
            raise ValueError("DoubleCrossfitAIPTW requires that n_splits >= 3")

        # Storing some information
        self._n_splits_ = n_splits
        self._n_partitions = n_partitions
        self._combine_method_ = method

        # Creating blank lists
        diff_est, diff_var, ratio_est, ratio_var = [], [], [], []

        # Conducts the re-sampling procedure
        if random_state is None:
            random_state = [None] * n_partitions
        else:
            random_state = RandomState(random_state).choice(range(5000000), size=n_partitions, replace=False)
        for j in range(self._n_partitions):
            # Estimating for a particular split (lots of functions happening in the background)
            result = self._single_crossfit_(random_state=random_state[j])

            # Appending results of this particular split combination
            diff_est.append(result[0])
            diff_var.append(result[1])
            if not self._continuous_outcome_:
                ratio_est.append(result[2])
                ratio_var.append(result[3])

        # Obtaining overall estimate and (1-alpha)% CL from all splits
        zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        est, var = calculate_joint_estimate(diff_est, diff_var, method=method)
        if self._continuous_outcome_:
            self.ace_vector = diff_est
            self.ace_var_vector = diff_var
            self.ace = est
            self.ace_se = np.sqrt(var)
            self.ace_ci = (self.ace - zalpha*self.ace_se,
                           self.ace + zalpha*self.ace_se)
        else:
            # Risk Difference
            self.risk_difference_vector = diff_est
            self.risk_difference_var_vector = diff_var
            self.risk_difference = est
            self.risk_difference_se = np.sqrt(var)
            self.risk_difference_ci = (self.risk_difference - zalpha*self.risk_difference_se,
                                       self.risk_difference + zalpha*self.risk_difference_se)
            # Risk Ratio
            self.risk_ratio_vector = ratio_est
            self.risk_ratio_var_vector = ratio_var
            ln_rr, ln_rr_var = calculate_joint_estimate(np.log(self.risk_ratio_vector),
                                                        self.risk_ratio_var_vector, method=method)
            self.risk_ratio = np.exp(ln_rr)
            self.risk_ratio_se = np.sqrt(ln_rr_var)
            self.risk_ratio_ci = (np.exp(ln_rr - zalpha*self.risk_ratio_se),
                                  np.exp(ln_rr + zalpha*self.risk_ratio_se))

    def summary(self, decimal=3):
        """Prints summary of model results

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if (self._fit_outcome_ is False) or (self._fit_treatment_ is False):
            raise ValueError('exposure_model and outcome_model must be specified before the estimate can '
                             'be generated')

        print('======================================================================')
        print('                     Double Cross-fit AIPTW                           ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.exposure, self.df.shape[0]))
        fmt = 'Outcome:          {:<15} No. of Splits:        {:<20}'
        print(fmt.format(self.outcome, self._n_splits_))
        fmt = 'Method:           {:<15} No. of Partitions:    {:<20}'
        print(fmt.format(self._combine_method_, self._n_partitions))

        print('======================================================================')
        if self._continuous_outcome_:
            print('Average Causal Effect: ', round(float(self.ace), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.ace_ci[0], decimal)), ',',
                  str(round(self.ace_ci[1], decimal)) + ')')
        else:
            print('Risk Difference:    ', round(float(self.risk_difference), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_difference_ci[0], decimal)), ',',
                  str(round(self.risk_difference_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Risk Ratio:         ', round(float(self.risk_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_ratio_ci[0], decimal)), ',',
                  str(round(self.risk_ratio_ci[1], decimal)) + ')')

        print('======================================================================')

    def run_diagnostics(self, color='gray'):
        """Runs available diagnostics for the plots. Currently diagnostics consist of a plot of the different point
        estimates and variance estimates across different partitions. Diagnostics for cross-fit estimators is ongoing.
        If you have any suggestions, please feel free to contact me on GitHub

        Parameters
        ----------
        color : str, optional
            Controls color of the plots. Default is gray

        Returns
        -------
        Plot to console
        """
        # Continuous outcomes have less plots to generate
        if self._continuous_outcome_:
            _run_diagnostic_(diff=self.ace_vector, diff_var=self.ace_var_vector,
                             color=color)

        # Binary outcomes have plots for all measures
        else:
            _run_diagnostic_(diff=self.risk_difference_vector, diff_var=self.risk_difference_var_vector,
                             rratio=self.risk_ratio_vector, rratio_var=self.risk_ratio_var_vector,
                             color=color)

    def _single_crossfit_(self, random_state):
        """Background function that runs a single crossfit of the split samples
        """
        # Dividing into s different splits
        sample_split = _sample_split_(self.df, n_splits=self._n_splits_, random_state=random_state)

        # Determining pairings to use for each sample split and each combination
        pairing_exposure = [i - 1 for i in range(self._n_splits_)]
        pairing_outcome = [i - 2 for i in range(self._n_splits_)]

        # Estimating treatment nuisance model
        a_models = _treatment_nuisance_(treatment=self.exposure, estimator=self._a_estimator,
                                        samples=sample_split, covariates=self._a_covariates)
        # Estimating outcome nuisance model
        y_models = _outcome_nuisance_(outcome=self.outcome, estimator=self._y_estimator,
                                      samples=sample_split, covariates=self._y_covariates)

        # Generating predictions based on set pairs for cross-fit procedure
        predictions = []
        y_obs, a_obs = np.array([]), np.array([])
        split_index = []
        for id, ep, op in zip(range(self._n_splits_), pairing_exposure, pairing_outcome):
            predictions.append(self._generate_predictions_(sample_split[id],
                                                           a_model_v=a_models[ep],
                                                           y_model_v=y_models[op]))
            # Generating vector of Y in correct order
            y_obs = np.append(y_obs, np.asarray(sample_split[id][self.outcome]))
            # Generating vector of A in correct order
            a_obs = np.append(a_obs, np.asarray(sample_split[id][self.exposure]))
            # Generating index for splits
            split_index.extend([id]*sample_split[id].shape[0])

        # Stacking Predicted Pr(A=1), Y(a=1), Y(a=0)
        pred_a_array, pred_y1_array, pred_y0_array = np.array([]), np.array([]), np.array([])
        for preds in predictions:
            pred_a_array = np.append(pred_a_array, preds[0])
            pred_y1_array = np.append(pred_y1_array, preds[1])
            pred_y0_array = np.append(pred_y0_array, preds[2])

        # Applying bounds if requested
        if self._gbounds:  # Bounding g-model if requested
            pred_a_array = probability_bounds(pred_a_array, bounds=self._gbounds)

        # Calculating point estimates
        difference, var_diff = aipw_calculator(y=y_obs, a=a_obs,
                                               py_a=pred_y1_array, py_n=pred_y0_array,
                                               pa1=pred_a_array, pa0=1-pred_a_array,
                                               splits=np.asarray(split_index),
                                               difference=True, continuous=self._continuous_outcome_)
        if self._continuous_outcome_:
            return difference, var_diff
        else:
            ratio, var_ratio = aipw_calculator(y=y_obs, a=a_obs,
                                               py_a=pred_y1_array, py_n=pred_y0_array,
                                               pa1=pred_a_array, pa0=1 - pred_a_array,
                                               splits=np.asarray(split_index),
                                               difference=False, continuous=False)
            return difference, var_diff, ratio, var_ratio

    def _generate_predictions_(self, sample, a_model_v, y_model_v):
        """Generates predictions from fitted functions (in background of _single_crossfit()
        """
        s = sample.copy()

        # Predicting Pr(A=1|L)
        xdata = np.asarray(patsy.dmatrix(self._a_covariates + ' - 1', s))
        a_pred = _ml_predictor(xdata, fitted_algorithm=a_model_v)

        # Predicting E(Y|A=1, L)
        s[self.exposure] = 1
        xdata = np.asarray(patsy.dmatrix(self._y_covariates + ' - 1', s))
        y_treat = _ml_predictor(xdata, fitted_algorithm=y_model_v)

        # Predicting E(Y|A=0, L)
        s[self.exposure] = 0
        xdata = np.asarray(patsy.dmatrix(self._y_covariates + ' - 1', s))
        y_none = _ml_predictor(xdata, fitted_algorithm=y_model_v)

        return a_pred, y_treat, y_none


class SingleCrossfitTMLE:
    """Implementation of the Targeted Maximum Likelihood Estimator with a single cross-fit procedure. The purpose
    of the cross-fit procedure is to all for non-Donsker nuisance function estimators. Some of machine learning
    algorithms are non-Donsker. In practice this means that confidence interval coverage can be incorrect when certain
    nuisance function estimators are used. Additionally, bias may persist as well. Cross-fitting is meant to alleviate
    this issue, therefore cross-fitting with a doubly-robust estimator is recommended when using machine learning.

    `SingleCrossfitTMLE` uses a single cross-fit, where the data set is paritioned into at least two non-overlapping
    splits. The nuisance function estimators are then estimated in each split. The estimated nuisance functions are
    then used to predict values in a non-overlapping split. This decouple the nuisance function estimation from the
    data used to estimate it

    Note
    ----
    Because of the repetitions of the procedure are needed to reduce variance determined by a particular partition, it
    can take a long time to run this code.

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing all necessary variables
    exposure : str
        Label for treatment column in the pandas data frame
    outcome : str
        Label for outcome column in the pandas data frame
    alpha : float, optional
        Alpha for confidence interval level. Default is 0.05
    continuous_bound : float, optional
        Optional argument to control the bounding feature for continuous outcomes. The bounding process may result
        in values of 0,1 which are undefined for logit(x). This parameter adds or substracts from the scenarios of
        0,1 respectively. Default value is 0.0005

    Examples
    --------
    Setting up environment

    >>> from sklearn.linear_model import LogisticRegression
    >>> from zepid import load_sample_data
    >>> from zepid.causal.doublyrobust import SingleCrossfitTMLE
    >>> df = load_sample_data(False).drop(columns='cd4_wk45').dropna()

    Estimating the single cross-fit TMLE

    >>> sctmle = SingleCrossfitTMLE(df, exposure='art', outcome='dead')
    >>> sctmle.exposure_model("male + age0 + cd40 + dvl0", estimator=LogisticRegression(solver='lbfgs'))
    >>> sctmle.outcome_model("art + male + age0 + cd40 + dvl0", estimator=LogisticRegression(solver='lbfgs'))
    >>> sctmle.fit(n_splits=5, n_partitions=100)
    >>> sctmle.summary()

    References
    ----------
    Chernozhukov V, Chetverikov D, Demirer M, Duflo E, Hansen C, Newey W, & Robins J. (2018). "Double/debiased machine
    learning for treatment and structural parameters". The Econometrics Journal 21:1; pC1–C6
    """
    def __init__(self, df, exposure, outcome, alpha=0.05, continuous_bound=0.0005):
        self.exposure = exposure
        self.outcome = outcome
        self.df, self._miss_flag, self._continuous_outcome_ = check_input_data(data=df,
                                                                               exposure=exposure,
                                                                               outcome=outcome,
                                                                               estimator="SingleCrossfitTMLE",
                                                                               drop_censoring=True,
                                                                               drop_missing=True,
                                                                               binary_exposure_only=True)
        self.alpha = alpha

        # bounding for continuous Y
        if self._continuous_outcome_:
            self._continuous_min = np.min(self.df[outcome])
            self._continuous_max = np.max(self.df[outcome])
            self._cb = continuous_bound
            self.df[outcome] = tmle_unit_bounds(y=self.df[outcome], mini=self._continuous_min,
                                                maxi=self._continuous_max, bound=self._cb)
        else:
            self._cb = 0.0

        self._a_covariates = None
        self._y_covariates = None
        self._a_estimator = None
        self._y_estimator = None
        self._fit_treatment_ = False
        self._fit_outcome_ = False
        self._gbounds = None
        self._n_splits_ = 0
        self._n_partitions = 0
        self._combine_method_ = None

        self.ace_vector = None
        self.ace_var_vector = None
        self.ace = None
        self.ace_ci = None
        self.ace_se = None

        self.risk_difference_vector = None
        self.risk_difference_var_vector = None
        self.risk_difference = None
        self.risk_difference_ci = None
        self.risk_difference_se = None

        self.risk_ratio_vector = None
        self.risk_ratio_var_vector = None
        self.risk_ratio = None
        self.risk_ratio_ci = None
        self.risk_ratio_se = None

        self.odds_ratio_vector = None
        self.odds_ratio_var_vector = None
        self.odds_ratio = None
        self.odds_ratio_se = None
        self.odds_ratio_ci = None

    def exposure_model(self, covariates, estimator, bound=False):
        """Specify the treatment nuisance model variables and estimator(s) to use. These parameters are held
        in the background until the .fit() function is called. These approaches are for used each sample split

        Parameters
        ----------
        covariates : str
            Confounders to include in the propensity score model. Follows patsy notation
        estimator :
            Estimator to use for prediction of the propensity score
        bound : float, list, optional
            Whether to bound predicted probabilities. Default is False, which does not bound
        """
        self._a_estimator = estimator
        self._a_covariates = covariates
        self._fit_treatment_ = True
        self._gbounds = bound

    def outcome_model(self, covariates, estimator):
        """Specify the outcome nuisance model variables and estimator(s) to use. These parameters are held
        in the background until the .fit() function is called. These approaches are for used each sample split

        Parameters
        ----------
        covariates : str
            Confounders to include in the propensity score model. Follows patsy notation
        estimator :
            Estimator to use for prediction of the propensity score
        """
        self._y_estimator = estimator
        self._y_covariates = covariates
        self._fit_outcome_ = True

    def fit(self, n_splits=2, n_partitions=100, method='median', random_state=None):
        """Runs the crossfit estimation procedure with the targeted maximum likelihood estimator. The estimation
        process is completed for multiple different splits during the procedure. The final estimate is defined as
        either the median or mean of the causal measure from each of the different splits. Median is used as the
        default since it is more stable.

        Note
        ----
        `n_partition` should be kept high to reduce dependency of results on the chosen number of splits

        Confidence intervals come from influences curves and incorporates the within-split variance and between-split
        variance.

        Parameters
        ----------
        n_splits : int
            Number of splits to use with a default of 2. The number of splits must be greater than or equal to 2.
        n_partitions : int
            Number of times to repeat the partition process. The default is 100, which I have seen good performance
            with in the past. Note that this algorithm can take a long time to run for high values of this parameter.
            It is best to test out run-times on small numbers first. Also if running in parallel, it can be reduced
        method : str, optional
            Method to obtain point estimates and standard errors. Median method takes the median (which is more robust)
            and the mean takes the mean. It has been remarked that the median is preferred, since it is more stable to
            extreme outliers, which may happen in finite samples
        random_state : None, int, optional
            Whether to set a seed for the partitions. Default is None (which does not use a user-set seed). Any valid
            NumPy seed can be input. Note that you should also state the random_state of all (applicable) estimators
            to ensure replicability. Seeds are chosen by the following procedure. The input random_state is based to
            np.random.choice to select n_partitions between 0 and 5million. That list of n_partition-length is then
            passed to each iteration of the cross-fitting pandas.DataFrame.sample(random_state).
        """
        # Checking for various issues
        if not self._fit_treatment_:
            raise ValueError("exposure_model() must be called before fit()")
        if not self._fit_outcome_:
            raise ValueError("outcome_model() must be called before fit()")
        if n_splits < 2:
            raise ValueError("SingleCrossfitTMLE requires that n_splits >= 2")

        # Storing some information
        self._n_splits_ = n_splits
        self._n_partitions = n_partitions
        self._combine_method_ = method

        # Creating blank lists
        diff_est, diff_var, rratio_est, rratio_var, oratio_est, oratio_var = [], [], [], [], [], []

        # Conducts the re-sampling procedure
        if random_state is None:
            random_state = [None] * n_partitions
        else:
            random_state = RandomState(random_state).choice(range(5000000), size=n_partitions, replace=False)
        for j in range(self._n_partitions):
            # Estimating for a particular split (lots of functions happening in the background)
            result = self._single_crossfit_(random_state=random_state[j])

            # Appending results of this particular split combination
            diff_est.append(result[0])
            diff_var.append(result[1])
            if not self._continuous_outcome_:
                rratio_est.append(result[2])
                rratio_var.append(result[3])
                oratio_est.append(result[4])
                oratio_var.append(result[5])

        # Obtaining overall estimate and (1-alpha)% CL from all splits
        zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        est, var = calculate_joint_estimate(diff_est, diff_var, method=method)
        if self._continuous_outcome_:
            self.ace_vector = diff_est
            self.ace_var_vector = diff_var
            self.ace = est
            self.ace_se = np.sqrt(var)
            self.ace_ci = (self.ace - zalpha*self.ace_se,
                           self.ace + zalpha*self.ace_se)
        else:
            # Risk Difference
            self.risk_difference_vector = diff_est
            self.risk_difference_var_vector = diff_var
            self.risk_difference = est
            self.risk_difference_se = np.sqrt(var)
            self.risk_difference_ci = (self.risk_difference - zalpha*self.risk_difference_se,
                                       self.risk_difference + zalpha*self.risk_difference_se)
            # Risk Ratio
            self.risk_ratio_vector = rratio_est
            self.risk_ratio_var_vector = rratio_var
            ln_rr, ln_rr_var = calculate_joint_estimate(np.log(self.risk_ratio_vector),
                                                        self.risk_ratio_var_vector, method=method)
            self.risk_ratio = np.exp(ln_rr)
            self.risk_ratio_se = np.sqrt(ln_rr_var)
            self.risk_ratio_ci = (np.exp(ln_rr - zalpha*self.risk_ratio_se),
                                  np.exp(ln_rr + zalpha*self.risk_ratio_se))
            # Odds Ratio
            self.odds_ratio_vector = oratio_est
            self.odds_ratio_var_vector = oratio_var
            ln_or, ln_or_var = calculate_joint_estimate(np.log(self.odds_ratio_vector),
                                                        self.odds_ratio_var_vector, method=method)
            self.odds_ratio = np.exp(ln_or)
            self.odds_ratio_se = np.sqrt(ln_or_var)
            self.odds_ratio_ci = (np.exp(ln_or - zalpha*self.odds_ratio_se),
                                  np.exp(ln_or + zalpha*self.odds_ratio_se))

    def summary(self, decimal=3):
        """Prints summary of model results

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if (self._fit_outcome_ is False) or (self._fit_treatment_ is False):
            raise ValueError('exposure_model and outcome_model must be specified before the estimate can '
                             'be generated')

        print('======================================================================')
        print('                     Single Cross-fit TMLE                            ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.exposure, self.df.shape[0]))
        fmt = 'Outcome:          {:<15} No. of Splits:        {:<20}'
        print(fmt.format(self.outcome, self._n_splits_))
        fmt = 'Method:           {:<15} No. of Partitions:    {:<20}'
        print(fmt.format(self._combine_method_, self._n_partitions))

        print('======================================================================')
        if self._continuous_outcome_:
            print('Average Causal Effect: ', round(float(self.ace), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.ace_ci[0], decimal)), ',',
                  str(round(self.ace_ci[1], decimal)) + ')')
        else:
            print('Risk Difference:    ', round(float(self.risk_difference), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_difference_ci[0], decimal)), ',',
                  str(round(self.risk_difference_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Risk Ratio:         ', round(float(self.risk_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_ratio_ci[0], decimal)), ',',
                  str(round(self.risk_ratio_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Odds Ratio:         ', round(float(self.odds_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.odds_ratio_ci[0], decimal)), ',',
                  str(round(self.odds_ratio_ci[1], decimal)) + ')')
        print('======================================================================')

    def run_diagnostics(self, color='gray'):
        """Runs available diagnostics for the plots. Currently diagnostics consist of a plot of the different point
        estimates and variance estimates across different partitions. Diagnostics for cross-fit estimators is ongoing.
        If you have any suggestions, please feel free to contact me on GitHub

        Parameters
        ----------
        color : str, optional
            Controls color of the plots. Default is gray

        Returns
        -------
        Plot to console
        """
        # Continuous outcomes have less plots to generate
        if self._continuous_outcome_:
            _run_diagnostic_(diff=self.ace_vector, diff_var=self.ace_var_vector,
                             color=color)

        # Binary outcomes have plots for all measures
        else:
            _run_diagnostic_(diff=self.risk_difference_vector, diff_var=self.risk_difference_var_vector,
                             rratio=self.risk_ratio_vector, rratio_var=self.risk_ratio_var_vector,
                             oratio=self.odds_ratio_vector, oratio_var=self.odds_ratio_var_vector,
                             color=color)

    def _single_crossfit_(self, random_state):
        """Background function that runs a single crossfit of the split samples
        """
        # Dividing into s different splits
        sample_split = _sample_split_(self.df, n_splits=self._n_splits_, random_state=random_state)

        # Determining pairings to use for each sample split and each combination
        pairing_exposure = [i - 1 for i in range(self._n_splits_)]
        pairing_outcome = pairing_exposure

        # Estimating treatment nuisance model
        a_models = _treatment_nuisance_(treatment=self.exposure, estimator=self._a_estimator,
                                        samples=sample_split, covariates=self._a_covariates)
        # Estimating outcome nuisance model
        y_models = _outcome_nuisance_(outcome=self.outcome, estimator=self._y_estimator,
                                      samples=sample_split, covariates=self._y_covariates)

        # Generating predictions based on set pairs for cross-fit procedure
        predictions = []
        y_obs, a_obs = np.array([]), np.array([])
        split_index = []
        for id, ep, op in zip(range(self._n_splits_), pairing_exposure, pairing_outcome):
            predictions.append(self._generate_predictions_(sample_split[id],
                                                           a_model_v=a_models[ep],
                                                           y_model_v=y_models[op]))
            # Generating vector of Y in correct order
            y_obs = np.append(y_obs, np.asarray(sample_split[id][self.outcome]))
            # Generating vector of A in correct order
            a_obs = np.append(a_obs, np.asarray(sample_split[id][self.exposure]))
            # Generating index for splits
            split_index.extend([id]*sample_split[id].shape[0])

        # Stacking Predicted Pr(A=1), Y(a=1), Y(a=0)
        pred_a_array, pred_y1_array, pred_y0_array = np.array([]), np.array([]), np.array([])
        for preds in predictions:
            pred_a_array = np.append(pred_a_array, preds[0])
            pred_y1_array = np.append(pred_y1_array, preds[1])
            pred_y0_array = np.append(pred_y0_array, preds[2])

        # Applying bounds if requested
        if self._gbounds:  # Bounding g-model if requested
            pred_a_array = probability_bounds(pred_a_array, bounds=self._gbounds)

        # Calculating point estimates
        targeted_vals = targeting_step(y=y_obs, a=a_obs,
                                       py_a=pred_y1_array, py_n=pred_y0_array,
                                       pa1=pred_a_array, pa0=1-pred_a_array,
                                       splits=np.asarray(split_index))

        if self._continuous_outcome_:
            difference, var_diff = tmle_calculator(y=y_obs,
                                                   ystar1=targeted_vals[0], ystar0=targeted_vals[1],
                                                   ystara=targeted_vals[2],
                                                   h1w=targeted_vals[3], h0w=targeted_vals[4], haw=targeted_vals[5],
                                                   splits=np.asarray(split_index),
                                                   measure='ate',
                                                   lower_bound=self._continuous_min, upper_bound=self._continuous_max)
            return difference, var_diff
        else:
            difference, var_diff = tmle_calculator(y=y_obs,
                                                   ystar1=targeted_vals[0], ystar0=targeted_vals[1],
                                                   ystara=targeted_vals[2],
                                                   h1w=targeted_vals[3], h0w=targeted_vals[4], haw=targeted_vals[5],
                                                   splits=np.asarray(split_index),
                                                   measure='risk_difference')
            rratio, var_rratio = tmle_calculator(y=y_obs,
                                                 ystar1=targeted_vals[0], ystar0=targeted_vals[1],
                                                 ystara=targeted_vals[2],
                                                 h1w=targeted_vals[3], h0w=targeted_vals[4], haw=targeted_vals[5],
                                                 splits=np.asarray(split_index),
                                                 measure='risk_ratio')
            oratio, var_oratio = tmle_calculator(y=y_obs,
                                                 ystar1=targeted_vals[0], ystar0=targeted_vals[1],
                                                 ystara=targeted_vals[2],
                                                 h1w=targeted_vals[3], h0w=targeted_vals[4], haw=targeted_vals[5],
                                                 splits=np.asarray(split_index),
                                                 measure='odds_ratio')
            return difference, var_diff, rratio, var_rratio, oratio, var_oratio

    def _generate_predictions_(self, sample, a_model_v, y_model_v):
        """Generates predictions from fitted functions (in background of _single_crossfit()
        """
        s = sample.copy()

        # Predicting Pr(A=1|L)
        xdata = np.asarray(patsy.dmatrix(self._a_covariates + ' - 1', s))
        a_pred = _ml_predictor(xdata, fitted_algorithm=a_model_v)

        # Predicting E(Y|A=1, L)
        s[self.exposure] = 1
        xdata = np.asarray(patsy.dmatrix(self._y_covariates + ' - 1', s))
        y_treat = _ml_predictor(xdata, fitted_algorithm=y_model_v)

        # Predicting E(Y|A=0, L)
        s[self.exposure] = 0
        xdata = np.asarray(patsy.dmatrix(self._y_covariates + ' - 1', s))
        y_none = _ml_predictor(xdata, fitted_algorithm=y_model_v)

        return a_pred, y_treat, y_none


class DoubleCrossfitTMLE:
    """Implementation of the Targeted Maximum Likelihood Estimator with a double cross-fit procedure. The purpose
    of the cross-fit procedure is to all for non-Donsker nuisance function estimators. Some of machine learning
    algorithms are non-Donsker. In practice this means that confidence interval coverage can be incorrect when certain
    nuisance function estimators are used. Additionally, bias may persist as well. Cross-fitting is meant to alleviate
    this issue, therefore cross-fitting with a doubly-robust estimator is recommended when using machine learning.

    `DoubleCrossfitTMLE` uses a double cross-fit, where the data set is paritioned into at least three non-overlapping
    split. The nuisance function estimators are then estimated in each split. The estimated nuisance functions are
    then used to predict values in a non-overlapping split. This decouple the nuisance function estimation from the
    data used to estimate it

    Note
    ----
    Because of the repetitions of the procedure are needed to reduce variance determined by a particular partition, it
    can take a long time to run this code.

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing all necessary variables
    exposure : str
        Label for treatment column in the pandas data frame
    outcome : str
        Label for outcome column in the pandas data frame
    alpha : float, optional
        Alpha for confidence interval level. Default is 0.05
    continuous_bound : float, optional
        Optional argument to control the bounding feature for continuous outcomes. The bounding process may result
        in values of 0,1 which are undefined for logit(x). This parameter adds or substracts from the scenarios of
        0,1 respectively. Default value is 0.0005

    Examples
    --------
    Setting up environment

    >>> from sklearn.linear_model import LogisticRegression
    >>> from zepid import load_sample_data
    >>> from zepid.causal.doublyrobust import DoubleCrossfitTMLE
    >>> df = load_sample_data(False).drop(columns='cd4_wk45').dropna()

    Estimating the double cross-fit TMLE

    >>> dctmle = DoubleCrossfitTMLE(df, exposure='art', outcome='dead')
    >>> dctmle.exposure_model("male + age0 + cd40 + dvl0", estimator=LogisticRegression(solver='lbfgs'))
    >>> dctmle.outcome_model("art + male + age0 + cd40 + dvl0", estimator=LogisticRegression(solver='lbfgs'))
    >>> dctmle.fit(n_splits=5, n_partitions=100)
    >>> dctmle.summary()

    References
    ----------
    Zivich PN, & Breskin A. (2020). Machine learning for causal inference: on the use of cross-fit estimators.
    arXiv preprint arXiv:2004.10337.

    Newey WK, Robins JR. (2018) "Cross-fitting and fast remainder rates for semiparametric estimation".
    arXiv:1801.09138

    Chernozhukov V, Chetverikov D, Demirer M, Duflo E, Hansen C, Newey W, & Robins J. (2018). "Double/debiased machine
    learning for treatment and structural parameters". The Econometrics Journal 21:1; pC1–C6
    """
    def __init__(self, df, exposure, outcome, alpha=0.05, continuous_bound=0.0005):
        self.exposure = exposure
        self.outcome = outcome
        self.df, self._miss_flag, self._continuous_outcome_ = check_input_data(data=df,
                                                                               exposure=exposure,
                                                                               outcome=outcome,
                                                                               estimator="DoubleCrossfitTMLE",
                                                                               drop_censoring=True,
                                                                               drop_missing=True,
                                                                               binary_exposure_only=True)
        self.alpha = alpha

        # bounding for continuous Y
        if self._continuous_outcome_:
            self._continuous_min = np.min(self.df[outcome])
            self._continuous_max = np.max(self.df[outcome])
            self._cb = continuous_bound
            self.df[outcome] = tmle_unit_bounds(y=self.df[outcome], mini=self._continuous_min,
                                                maxi=self._continuous_max, bound=self._cb)
        else:
            self._cb = 0.0

        self._a_covariates = None
        self._y_covariates = None
        self._a_estimator = None
        self._y_estimator = None
        self._fit_treatment_ = False
        self._fit_outcome_ = False
        self._gbounds = None
        self._n_splits_ = 0
        self._n_partitions = 0
        self._combine_method_ = None

        self.ace_vector = None
        self.ace_var_vector = None
        self.ace = None
        self.ace_ci = None
        self.ace_se = None

        self.risk_difference_vector = None
        self.risk_difference_var_vector = None
        self.risk_difference = None
        self.risk_difference_ci = None
        self.risk_difference_se = None

        self.risk_ratio_vector = None
        self.risk_ratio_var_vector = None
        self.risk_ratio = None
        self.risk_ratio_ci = None
        self.risk_ratio_se = None

        self.odds_ratio_vector = None
        self.odds_ratio_var_vector = None
        self.odds_ratio = None
        self.odds_ratio_se = None
        self.odds_ratio_ci = None

    def exposure_model(self, covariates, estimator, bound=False):
        """Specify the treatment nuisance model variables and estimator(s) to use. These parameters are held
        in the background until the .fit() function is called. These approaches are for used each sample split

        Parameters
        ----------
        covariates : str
            Confounders to include in the propensity score model. Follows patsy notation
        estimator :
            Estimator to use for prediction of the propensity score
        bound : float, list, optional
            Whether to bound predicted probabilities. Default is False, which does not bound
        """
        self._a_estimator = estimator
        self._a_covariates = covariates
        self._fit_treatment_ = True
        self._gbounds = bound

    def outcome_model(self, covariates, estimator):
        """Specify the outcome nuisance model variables and estimator(s) to use. These parameters are held
        in the background until the .fit() function is called. These approaches are for used each sample split

        Parameters
        ----------
        covariates : str
            Confounders to include in the propensity score model. Follows patsy notation
        estimator :
            Estimator to use for prediction of the propensity score
        """
        self._y_estimator = estimator
        self._y_covariates = covariates
        self._fit_outcome_ = True

    def fit(self, n_splits=3, n_partitions=100, method='median', random_state=None):
        """Runs the crossfit estimation procedure with the targeted maximum likelihood estimator. The
        estimation process is completed for multiple different splits during the procedure. The final estimate is
        defined as either the median or mean of the causal measure from each of the different splits. Median is
        used as the default since it is more stable.

        Note
        ----
        `n_partition` should be kept high to reduce dependency of results on the chosen number of splits

        Confidence intervals come from influences curves and incorporates the within-split variance and between-split
        variance.

        Parameters
        ----------
        n_splits : int
            Number of splits to use with a default of 3. The number of splits must be greater than or equal to 3.
        n_partitions : int
            Number of times to repeat the partition process. The default is 100, which I have seen good performance
            with in the past. Note that this algorithm can take a long time to run for high values of this parameter.
            It is best to test out run-times on small numbers first. Also if running in parallel, it can be reduced
        method : str, optional
            Method to obtain point estimates and standard errors. Median method takes the median (which is more robust)
            and the mean takes the mean. It has been remarked that the median is preferred, since it is more stable to
            extreme outliers, which may happen in finite samples
        random_state : None, int, optional
            Whether to set a seed for the partitions. Default is None (which does not use a user-set seed). Any valid
            NumPy seed can be input. Note that you should also state the random_state of all (applicable) estimators
            to ensure replicability. Seeds are chosen by the following procedure. The input random_state is based to
            np.random.choice to select n_partitions between 0 and 5million. That list of n_partition-length is then
            passed to each iteration of the cross-fitting pandas.DataFrame.sample(random_state).
        """
        # Checking for various issues
        if not self._fit_treatment_:
            raise ValueError("exposure_model() must be called before fit()")
        if not self._fit_outcome_:
            raise ValueError("outcome_model() must be called before fit()")
        if n_splits < 3:
            raise ValueError("DoubleCrossfitTMLE requires that n_splits >= 3")

        # Storing some information
        self._n_splits_ = n_splits
        self._n_partitions = n_partitions
        self._combine_method_ = method

        # Creating blank lists
        diff_est, diff_var, rratio_est, rratio_var, oratio_est, oratio_var = [], [], [], [], [], []

        # Conducts the re-sampling procedure
        if random_state is None:
            random_state = [None] * n_partitions
        else:
            random_state = RandomState(random_state).choice(range(5000000), size=n_partitions, replace=False)
        for j in range(self._n_partitions):
            # Estimating for a particular split (lots of functions happening in the background)
            result = self._single_crossfit_(random_state=random_state[j])

            # Appending results of this particular split combination
            diff_est.append(result[0])
            diff_var.append(result[1])
            if not self._continuous_outcome_:
                rratio_est.append(result[2])
                rratio_var.append(result[3])
                oratio_est.append(result[4])
                oratio_var.append(result[5])

        # Obtaining overall estimate and (1-alpha)% CL from all splits
        zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        est, var = calculate_joint_estimate(diff_est, diff_var, method=method)
        if self._continuous_outcome_:
            self.ace_vector = diff_est
            self.ace_var_vector = diff_var
            self.ace = est
            self.ace_se = np.sqrt(var)
            self.ace_ci = (self.ace - zalpha*self.ace_se,
                           self.ace + zalpha*self.ace_se)
        else:
            # Risk Difference
            self.risk_difference_vector = diff_est
            self.risk_difference_var_vector = diff_var
            self.risk_difference = est
            self.risk_difference_se = np.sqrt(var)
            self.risk_difference_ci = (self.risk_difference - zalpha*self.risk_difference_se,
                                       self.risk_difference + zalpha*self.risk_difference_se)
            # Risk Ratio
            self.risk_ratio_vector = rratio_est
            self.risk_ratio_var_vector = rratio_var
            ln_rr, ln_rr_var = calculate_joint_estimate(np.log(self.risk_ratio_vector),
                                                        self.risk_ratio_var_vector, method=method)
            self.risk_ratio = np.exp(ln_rr)
            self.risk_ratio_se = np.sqrt(ln_rr_var)
            self.risk_ratio_ci = (np.exp(ln_rr - zalpha*self.risk_ratio_se),
                                  np.exp(ln_rr + zalpha*self.risk_ratio_se))
            # Odds Ratio
            self.odds_ratio_vector = oratio_est
            self.odds_ratio_var_vector = oratio_var
            ln_or, ln_or_var = calculate_joint_estimate(np.log(self.odds_ratio_vector),
                                                        self.odds_ratio_var_vector, method=method)
            self.odds_ratio = np.exp(ln_or)
            self.odds_ratio_se = np.sqrt(ln_or_var)
            self.odds_ratio_ci = (np.exp(ln_or - zalpha*self.odds_ratio_se),
                                  np.exp(ln_or + zalpha*self.odds_ratio_se))

    def summary(self, decimal=3):
        """Prints summary of model results

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if (self._fit_outcome_ is False) or (self._fit_treatment_ is False):
            raise ValueError('exposure_model and outcome_model must be specified before the estimate can '
                             'be generated')

        print('======================================================================')
        print('                     Double Cross-fit TMLE                            ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.exposure, self.df.shape[0]))
        fmt = 'Outcome:          {:<15} No. of Splits:        {:<20}'
        print(fmt.format(self.outcome, self._n_splits_))
        fmt = 'Method:           {:<15} No. of Partitions:    {:<20}'
        print(fmt.format(self._combine_method_, self._n_partitions))

        print('======================================================================')
        if self._continuous_outcome_:
            print('Average Causal Effect: ', round(float(self.ace), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.ace_ci[0], decimal)), ',',
                  str(round(self.ace_ci[1], decimal)) + ')')
        else:
            print('Risk Difference:    ', round(float(self.risk_difference), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_difference_ci[0], decimal)), ',',
                  str(round(self.risk_difference_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Risk Ratio:         ', round(float(self.risk_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_ratio_ci[0], decimal)), ',',
                  str(round(self.risk_ratio_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Odds Ratio:         ', round(float(self.odds_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.odds_ratio_ci[0], decimal)), ',',
                  str(round(self.odds_ratio_ci[1], decimal)) + ')')
        print('======================================================================')

    def run_diagnostics(self, color='gray'):
        """Runs available diagnostics for the plots. Currently diagnostics consist of a plot of the different point
        estimates and variance estimates across different partitions. Diagnostics for cross-fit estimators is ongoing.
        If you have any suggestions, please feel free to contact me on GitHub

        Parameters
        ----------
        color : str, optional
            Controls color of the plots. Default is gray

        Returns
        -------
        Plot to console
        """
        # Continuous outcomes have less plots to generate
        if self._continuous_outcome_:
            _run_diagnostic_(diff=self.ace_vector, diff_var=self.ace_var_vector,
                             color=color)

        # Binary outcomes have plots for all measures
        else:
            _run_diagnostic_(diff=self.risk_difference_vector, diff_var=self.risk_difference_var_vector,
                             rratio=self.risk_ratio_vector, rratio_var=self.risk_ratio_var_vector,
                             oratio=self.odds_ratio_vector, oratio_var=self.odds_ratio_var_vector,
                             color=color)

    def _single_crossfit_(self, random_state):
        """Background function that runs a single crossfit of the split samples
        """
        # Dividing into s different splits
        sample_split = _sample_split_(self.df, n_splits=self._n_splits_, random_state=random_state)

        # Determining pairings to use for each sample split and each combination
        pairing_exposure = [i - 1 for i in range(self._n_splits_)]
        pairing_outcome = [i - 2 for i in range(self._n_splits_)]

        # Estimating treatment nuisance model
        a_models = _treatment_nuisance_(treatment=self.exposure, estimator=self._a_estimator,
                                        samples=sample_split, covariates=self._a_covariates)
        # Estimating outcome nuisance model
        y_models = _outcome_nuisance_(outcome=self.outcome, estimator=self._y_estimator,
                                      samples=sample_split, covariates=self._y_covariates)

        # Generating predictions based on set pairs for cross-fit procedure
        predictions = []
        y_obs, a_obs = np.array([]), np.array([])
        split_index = []
        for id, ep, op in zip(range(self._n_splits_), pairing_exposure, pairing_outcome):
            predictions.append(self._generate_predictions_(sample_split[id],
                                                           a_model_v=a_models[ep],
                                                           y_model_v=y_models[op]))
            # Generating vector of Y in correct order
            y_obs = np.append(y_obs, np.asarray(sample_split[id][self.outcome]))
            # Generating vector of A in correct order
            a_obs = np.append(a_obs, np.asarray(sample_split[id][self.exposure]))
            # Generating index for splits
            split_index.extend([id]*sample_split[id].shape[0])

        # Stacking Predicted Pr(A=1), Y(a=1), Y(a=0)
        pred_a_array, pred_y1_array, pred_y0_array = np.array([]), np.array([]), np.array([])
        for preds in predictions:
            pred_a_array = np.append(pred_a_array, preds[0])
            pred_y1_array = np.append(pred_y1_array, preds[1])
            pred_y0_array = np.append(pred_y0_array, preds[2])

        # Applying bounds if requested
        if self._gbounds:  # Bounding g-model if requested
            pred_a_array = probability_bounds(pred_a_array, bounds=self._gbounds)

        # Calculating point estimates
        targeted_vals = targeting_step(y=y_obs, a=a_obs,
                                       py_a=pred_y1_array, py_n=pred_y0_array,
                                       pa1=pred_a_array, pa0=1-pred_a_array,
                                       splits=np.asarray(split_index))

        if self._continuous_outcome_:
            difference, var_diff = tmle_calculator(y=y_obs,
                                                   ystar1=targeted_vals[0], ystar0=targeted_vals[1],
                                                   ystara=targeted_vals[2],
                                                   h1w=targeted_vals[3], h0w=targeted_vals[4], haw=targeted_vals[5],
                                                   splits=np.asarray(split_index),
                                                   measure='ate',
                                                   lower_bound=self._continuous_min, upper_bound=self._continuous_max)
            return difference, var_diff
        else:
            difference, var_diff = tmle_calculator(y=y_obs,
                                                   ystar1=targeted_vals[0], ystar0=targeted_vals[1],
                                                   ystara=targeted_vals[2],
                                                   h1w=targeted_vals[3], h0w=targeted_vals[4], haw=targeted_vals[5],
                                                   splits=np.asarray(split_index),
                                                   measure='risk_difference')
            rratio, var_rratio = tmle_calculator(y=y_obs,
                                                 ystar1=targeted_vals[0], ystar0=targeted_vals[1],
                                                 ystara=targeted_vals[2],
                                                 h1w=targeted_vals[3], h0w=targeted_vals[4], haw=targeted_vals[5],
                                                 splits=np.asarray(split_index),
                                                 measure='risk_ratio')
            oratio, var_oratio = tmle_calculator(y=y_obs,
                                                 ystar1=targeted_vals[0], ystar0=targeted_vals[1],
                                                 ystara=targeted_vals[2],
                                                 h1w=targeted_vals[3], h0w=targeted_vals[4], haw=targeted_vals[5],
                                                 splits=np.asarray(split_index),
                                                 measure='odds_ratio')
            return difference, var_diff, rratio, var_rratio, oratio, var_oratio

    def _generate_predictions_(self, sample, a_model_v, y_model_v):
        """Generates predictions from fitted functions (in background of _single_crossfit()
        """
        s = sample.copy()

        # Predicting Pr(A=1|L)
        xdata = np.asarray(patsy.dmatrix(self._a_covariates + ' - 1', s))
        a_pred = _ml_predictor(xdata, fitted_algorithm=a_model_v)

        # Predicting E(Y|A=1, L)
        s[self.exposure] = 1
        xdata = np.asarray(patsy.dmatrix(self._y_covariates + ' - 1', s))
        y_treat = _ml_predictor(xdata, fitted_algorithm=y_model_v)

        # Predicting E(Y|A=0, L)
        s[self.exposure] = 0
        xdata = np.asarray(patsy.dmatrix(self._y_covariates + ' - 1', s))
        y_none = _ml_predictor(xdata, fitted_algorithm=y_model_v)

        return a_pred, y_treat, y_none


def calculate_joint_estimate(point_est, var_est, method):
    """Function that combines the different estimates across partitions into a single point estimate and standard
    error. This function can be directly called and have data directly input into. Use of this function allows for
    users to run the cross-fit procedure in parallel, output the answers, and then produce the final point estimate
    via this function.

    Note
    ----
    This function is only intended to make running the parallel cross-fit procedure easier to implement for users.
    It does not actually run any part of the cross-fitting procedure. It only calculates a summary from disparate
    partitions

    Parameters
    ----------
    point_est : container
        Container of point estimates
    var_est : container
        Container of variance estimates
    method : string
        Method to combine results. Options are 'median' or 'mean'. Median is recommended since it is more stable
        across a fewer number of partitions
    """
    if len(point_est) != len(var_est):
        raise ValueError("The length of point_est and var_est do not match")

    # Using the Median Method
    if method == 'median':
        single_point = np.median(point_est)
        single_point_var = np.median(var_est + (point_est - single_point)**2)

    # Using the Mean Method
    elif method == 'mean':
        single_point = np.mean(point_est)
        single_point_var = np.mean(var_est + (point_est - single_point)**2)

    # Error if neither exists
    else:
        raise ValueError("Either 'mean' or 'median' must be selected for the pooling of repeated sample splits")

    return single_point, single_point_var


def targeting_step(y, a, py_a, py_n, pa1, pa0, splits):
    f = sm.families.family.Binomial()
    h1w = a / pa1
    h0w = -(1 - a) / pa0
    haw = h1w + h0w
    py_o = a * py_a + (1 - a) * py_n

    ystar1, ystar0, ystara = [], [], []
    for s in set(splits):
        ys = y[splits == s]
        pa1s = pa1[splits == s]
        pa0s = pa0[splits == s]
        py_as = py_a[splits == s]
        py_ns = py_n[splits == s]
        py_os = py_o[splits == s]
        h1ws = h1w[splits == s]
        h0ws = h0w[splits == s]

        # Targeting Step
        log = sm.GLM(ys, np.column_stack((h1ws, h0ws)), offset=np.log(probability_to_odds(py_os)),
                     family=f, missing='drop').fit()
        epsilon = log.params

        # Getting updated predictions from targeting step
        ystar1 = np.append(ystar1, logistic.cdf(np.log(probability_to_odds(py_as)) + epsilon[0] / pa1s))
        ystar0 = np.append(ystar0, logistic.cdf(np.log(probability_to_odds(py_ns)) - epsilon[1] / pa0s))
        ystara = np.append(ystara, log.predict(np.column_stack((h1ws, h0ws)),
                                               offset=np.log(probability_to_odds(py_os))))
    return ystar1, ystar0, ystara, h1w, h0w, haw


def tmle_calculator(y, ystar1, ystar0, ystara, h1w, h0w, haw, splits,
                    measure='ate', lower_bound=None, upper_bound=None):
    """Function to calculate TMLE estimates for SingleCrossfitTMLE, and DoubleCrossfitTMLE
    """
    if measure in ["ate", "risk_difference"]:
        # Unbounding if continuous outcome (ate)
        if measure == "ate":
            # Unbounding continuous outcomes
            y = tmle_unit_unbound(y, mini=lower_bound, maxi=upper_bound)
            ystar1 = tmle_unit_unbound(ystar1, mini=lower_bound, maxi=upper_bound)
            ystar0 = tmle_unit_unbound(ystar0, mini=lower_bound, maxi=upper_bound)
            ystara = tmle_unit_unbound(ystara, mini=lower_bound, maxi=upper_bound)

        # Point Estimate
        estimate = np.mean(ystar1 - ystar0)
        # Variance estimate
        variance = []
        for s in set(splits):
            ys = y[splits == s]
            ystar1s = ystar1[splits == s]
            ystar0s = ystar0[splits == s]
            ystaras = ystara[splits == s]
            haws = haw[splits == s]

            ic = haws * (ys - ystaras) + (ystar1s - ystar0s) - estimate
            variance.append(np.var(ic, ddof=1))

        return estimate, (np.mean(variance) / y.shape[0])

    elif measure == 'risk_ratio':
        # Point Estimate
        estimate = np.mean(ystar1) / np.mean(ystar0)
        variance = []
        for s in set(splits):
            ys = y[splits == s]
            ystar1s = ystar1[splits == s]
            ystar0s = ystar0[splits == s]
            ystaras = ystara[splits == s]
            h1ws = h1w[splits == s]
            h0ws = h0w[splits == s]

            ic = (1/np.mean(ystar1s) * (h1ws * (ys - ystaras)) + ystar1s - np.mean(ystar1s) -
                  (1/np.mean(ystar0s) * (-1 * h0ws * (ys - ystaras)) + ystar0s - np.mean(ystar0s)))
            variance.append(np.var(ic, ddof=1))

        return estimate, (np.mean(variance) / y.shape[0])

    elif measure == 'odds_ratio':
        # Point Estimate
        estimate = (np.mean(ystar1) / (1-np.mean(ystar1))) / (np.mean(ystar0) / (1-np.mean(ystar0)))
        variance = []
        for s in set(splits):
            ys = y[splits == s]
            ystar1s = ystar1[splits == s]
            ystar0s = ystar0[splits == s]
            ystaras = ystara[splits == s]
            h1ws = h1w[splits == s]
            h0ws = h0w[splits == s]

            ic = ((1-np.mean(ystar1s))/np.mean(ystar1s)*(h1ws*(ys - ystaras) + ystar1s) -
                  (1-np.mean(ystar0s))/np.mean(ystar0s)*(-1*h0ws*(ys - ystaras) + ystar0s))
            variance.append(np.var(ic, ddof=1))

        return estimate, (np.mean(variance) / y.shape[0])

    else:
        raise ValueError("Invalid measure requested within function: tmle_calculator. Input measure is " +
                         str(measure) + " but only 'ate', 'risk_difference', 'risk_ratio', and "
                                        "'odds_ratio' are accepted.")


def _sample_split_(data, n_splits, random_state=None):
    """Background function to split data into three non-overlapping pieces
    """
    # Break into approx even splits
    n = int(data.shape[0] / n_splits)

    splits = []
    data_to_sample = data.copy()
    # Procedures is done n_splits - 1 times
    for i in range(n_splits-1):  # Loops through splits and takes random sample all remaining sets of the data
        s = data_to_sample.sample(n=n, random_state=RandomState(random_state))
        splits.append(s.copy())
        data_to_sample = data_to_sample.loc[data_to_sample.index.difference(s.index)].copy()

    # Remaining data becomes last split
    splits.append(data_to_sample)
    return splits


def _ml_predictor(xdata, fitted_algorithm):
    """Background function to generate predictions of treatments
    """
    if hasattr(fitted_algorithm, 'predict_proba'):
        return fitted_algorithm.predict_proba(xdata)[:, 1]
    elif hasattr(fitted_algorithm, 'predict'):
        return fitted_algorithm.predict(xdata)


def _treatment_nuisance_(treatment, estimator, samples, covariates):
    """Procedure to fit the treatment ML
    """
    treatment_fit_splits = []
    for s in samples:
        # Using patsy to pull out the covariates
        xdata = np.asarray(patsy.dmatrix(covariates + ' - 1', s))
        ydata = np.asarray(s[treatment])

        # Fitting machine learner / super learner to each split
        est = copy.deepcopy(estimator)
        try:
            fm = est.fit(X=xdata, y=ydata)
            # print("Treatment model")
            # print(fm.summary())
        except TypeError:
            raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                            "covers both sklearn and supylearner")

        # Adding model to the list of models
        treatment_fit_splits.append(fm)

    return treatment_fit_splits


def _outcome_nuisance_(outcome, estimator, samples, covariates):
    """Background function to generate predictions of outcomes
    """
    outcome_fit_splits = []
    for s in samples:
        # Using patsy to pull out the covariates
        xdata = np.asarray(patsy.dmatrix(covariates + ' - 1', s))
        ydata = np.asarray(s[outcome])

        # Fitting machine learner / super learner to each
        est = copy.deepcopy(estimator)
        try:
            fm = est.fit(X=xdata, y=ydata)
            # print(est.summary())
        except TypeError:
            raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                            "covers both sklearn and supylearner")

        # Adding model to the list of models
        outcome_fit_splits.append(fm)

    return outcome_fit_splits


def _estimate_density_plot_(estimates, bw_method='scott', fill=True, color='gray', variance=False):
    """Generates a density plot of the different estimates for each of the different sample splits. Helps to visualize
    the variability between different splits. If there is high variability, this indicates there is high sensitivity
    to the particular chosen split.

    Returns
    -------

    """
    if variance:
        x = np.linspace(0, np.max(estimates)+0.005, 10000)
    else:
        x = np.linspace(np.min(estimates)-0.02, np.max(estimates)+0.02, 10000)
    density_t = gaussian_kde(estimates, bw_method=bw_method)

    # Plot
    ax = plt.gca()
    if fill:
        ax.fill_between(x, density_t(x), color=color, alpha=0.2, label=None)
    ax.plot(x, density_t(x), color=color)
    ax.set_yticks([])
    return ax


def _run_diagnostic_(diff, diff_var, rratio=None, rratio_var=None, oratio=None, oratio_var=None, color="gray"):
    """Background function to run all diagnostics

    Returns
    -------
    Plot to console
    """
    # Continuous outcomes have less plots to generate
    if rratio is None:
        # Point estimates
        plt.subplot(121)
        _estimate_density_plot_(diff, bw_method='scott', fill=True, color=color)
        plt.title("ACE")
        # Variance estimates
        plt.subplot(122)
        _estimate_density_plot_(diff_var, bw_method='scott', fill=True, color=color)
        plt.title("Var(ACE)")

    # Binary outcomes have plots for all measures
    else:
        if oratio is None:
            # Risk Difference estimates
            plt.subplot(221)
            _estimate_density_plot_(diff, bw_method='scott', fill=True, color=color)
            plt.title("Risk Difference")
            # Var(RD) estimates
            plt.subplot(223)
            _estimate_density_plot_(diff_var, bw_method='scott',
                                    fill=True, color=color, variance=True)
            plt.title("Var(RD)")

            # Risk Ratio estimates
            plt.subplot(222)
            _estimate_density_plot_(rratio, bw_method='scott', fill=True, color=color)
            plt.title("Risk Ratio")
            # Var(RR) estimates
            plt.subplot(224)
            _estimate_density_plot_(rratio_var, bw_method='scott',
                                    fill=True, color=color, variance=True)
            plt.title("Var(ln(RR))")
        else:
            # Risk Difference estimates
            plt.subplot(231)
            _estimate_density_plot_(diff, bw_method='scott', fill=True, color=color)
            plt.title("Risk Difference")
            # Var(RD) estimates
            plt.subplot(234)
            _estimate_density_plot_(diff_var, bw_method='scott',
                                    fill=True, color=color, variance=True)
            plt.title("Var(RD)")

            # Risk Ratio estimates
            plt.subplot(232)
            _estimate_density_plot_(rratio, bw_method='scott', fill=True, color=color)
            plt.title("Risk Ratio")
            # Var(RR) estimates
            plt.subplot(235)
            _estimate_density_plot_(rratio_var, bw_method='scott',
                                    fill=True, color=color, variance=True)
            plt.title("Var(ln(RR))")
            # Odds Ratio estimates
            plt.subplot(233)
            _estimate_density_plot_(oratio, bw_method='scott', fill=True, color=color)
            plt.title("Odds Ratio")
            # Var(OR) estimates
            plt.subplot(236)
            _estimate_density_plot_(oratio_var, bw_method='scott',
                                    fill=True, color=color, variance=True)
            plt.title("Var(ln(OR))")

    plt.tight_layout()
    plt.show()



from .graphics import (functional_form_plot,
                       EffectMeasurePlot,
                       pvalue_plot,
                       spaghetti_plot,
                       roc,
                       dynamic_risk_plot,
                       labbe_plot,
                       zipper_plot)



import warnings
import patsy
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from scipy.stats import logistic
from scipy.stats.kde import gaussian_kde
from statsmodels.stats.weightstats import DescrStatsW
import matplotlib.pyplot as plt

from zepid.calc import probability_to_odds, probability_bounds


def check_input_data(data, exposure, outcome, estimator, drop_censoring, drop_missing, binary_exposure_only):
    """Background function used by the various estimators to check the input data for possible issues or
    inconsistencies with the format expected by the estimator.

    Parameters
    ----------
    data : DataFrame
        Input dataframe to be processed
    exposure : str
        Exposure/treatment column name
    outcome : str
        Outcome column
    estimator : str
        Name of the input estimator. Reads back the estimator name in the warning system
    drop_censoring : bool
        Whether to drop censored observations. Some estimators do not handle censoring (so this is set to True)
    drop_missing : bool
        Argument currently does nothing. Will be utilized in the future (when some estimators support missing data
        aside from censoring)
    binary_exposure_only : bool
        Argument for whether only binary exposure / treatments are supported. This is currently True for most
        estimators

    Returns
    -------
    formatted DataFrame, boolean for missing data flag
    """
    # Handling missing data and censoring
    if drop_censoring:  # Censoring is not always supported by all estimators
        valid_obs = data.dropna(subset=[d for d in data.columns]).shape[0]
        if valid_obs != data.shape[0]:
            warnings.warn("There is missing data in the dataset. By default, " + str(estimator) +
                          " will drop all missing data (including missing outcome data). " + str(estimator) +
                          " will fit " + str(valid_obs) + ' of ' + str(data.shape[0]) + ' observations', UserWarning)
            data = data.copy().dropna().reset_index()
        else:
            data = data.copy().reset_index()
        miss_flag = False
        data['__missing_indicator__'] = 1
    else:
        valid_obs = data.dropna(subset=[d for d in data.columns if d != outcome]).shape[0]
        # Checking for other missing data
        if valid_obs != data.shape[0]:
            warnings.warn("There is missing data that is not the outcome in the data set. " + str(estimator) +
                          " will drop all missing data that is not missing outcome data. " + str(estimator) +
                          " will fit " + str(valid_obs) + " of " + str(data.shape[0]) + " observations", UserWarning)
            data = data.copy().dropna(subset=[d for d in data.columns if d != outcome]).reset_index()
        else:
            data = data.copy().reset_index()
        # Checking for censored data
        if valid_obs != data.dropna(subset=[outcome]).shape[0]:
            miss_flag = True
            data['__missing_indicator__'] = np.where(data[outcome].isna(), 0, 1)
        else:
            miss_flag = False
            data['__missing_indicator__'] = 1

    # Checking for only binary exposures
    if binary_exposure_only:
        if not data[exposure].value_counts().index.isin([0, 1]).all():
            raise ValueError(str(estimator) + " only supports binary exposures currently")

    # Check outcome data type
    if data[outcome].value_counts().index.isin([0, 1]).all():
        continuous = False
    else:
        continuous = True

    return data, miss_flag, continuous


def propensity_score(df, model, weights=None, print_results=True):
    """Generate propensity scores (probability) based on the model input. Uses logistic regression model
    to calculate

    Parameters
    -----------
    df : DataFrame
        Pandas Dataframe containing the variables of interest
    model : str
        Model to fit the logistic regression to. For example, 'y ~ var1 + var2'
    weights : str, optional
        Whether to estimate the model using weights. Default is None (unweighted)
    print_results : bool, optional
        Whether to print the logistic regression results. Default is True

    Returns
    -------------
    Fitted statsmodels GLM object

    Example
    ------------
    >>>from zepid import load_sample_data
    >>>from zepid.causal.utils import propensity_score
    >>>df = load_sample_data(timevary=False)
    >>>propensity_score(df=df,model='dead ~ art0 + male + dvl0')
    """
    f = sm.families.family.Binomial()
    if weights is None:
        log = smf.glm(model, df, family=f).fit()
    else:
        log = smf.glm(model, df, freq_weights=df[weights], family=f).fit()

    if print_results:
        print('==============================================================================')
        print('Propensity Score Model')
        print(log.summary())
        print('==============================================================================')
    return log


def exposure_machine_learner(xdata, ydata, ml_model, print_results=True):
    """Function to fit machine learning predictions. Used by TMLE to generate predicted probabilities of being
    treated (i.e. Pr(A=1 | L))
    """
    # Trying to fit the Machine Learning model
    try:
        fm = ml_model.fit(X=xdata, y=ydata)
    except TypeError:
        raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                        "covers both sklearn and supylearner. If there is a predictive model you would "
                        "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                        "can work on adding support")
    if print_results and hasattr(fm, 'summary'):  # SuPyLearner has a nice summarize function
        print('==============================================================================')
        print('Propensity Score Model')
        fm.summary()
        print('==============================================================================')

    # Generating predictions
    if hasattr(fm, 'predict_proba'):
        g = fm.predict_proba(xdata)
        # this allows support for pygam LogisticGAM, which only returns only 1 probability
        if g.ndim == 1:
            return g
        else:
            return g[:, 1]
    elif hasattr(fm, 'predict'):
        return fm.predict(xdata)
    else:
        raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")


def outcome_machine_learner(xdata, ydata, all_a, none_a, ml_model, continuous, print_results=True):
    """Function to fit machine learning predictions. Used by TMLE to generate predicted probabilities of outcome
    (i.e. Pr(Y=1 | A=1, L) and Pr(Y=1 | A=0, L)).
    """
    # Trying to fit Machine Learning model
    try:
        fm = ml_model.fit(X=xdata, y=ydata)
    except TypeError:
        raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                        "covers both sklearn and supylearner. If there is a predictive model you would "
                        "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                        "can work on adding support")
    if print_results and hasattr(fm, 'summary'):  # Nice summarize option from SuPyLearner
        print('==============================================================================')
        print('Outcome Model')
        fm.summary()
        print('==============================================================================')

    # Generating predictions
    if continuous:
        if hasattr(fm, 'predict'):
            qa1 = fm.predict(all_a)
            qa0 = fm.predict(none_a)
            return qa1, qa0
        else:
            raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")

    else:
        if hasattr(fm, 'predict_proba'):
            qa1 = fm.predict_proba(all_a)
            qa0 = fm.predict_proba(none_a)
            if (qa1.ndim == 1) and (qa0.ndim == 1):
                return qa1, qa0
            else:
                return qa1[:, 1], qa0[:, 1]
        elif hasattr(fm, 'predict'):
            qa1 = fm.predict(all_a)
            qa0 = fm.predict(none_a)
            return qa1, qa0
        else:
            raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")


def stochastic_outcome_machine_learner(xdata, ydata, ml_model, continuous, print_results=True):
    """Function to fit machine learning predictions. Used by StochasticTMLE to generate predicted probabilities of
    outcome (i.e. Pr(Y=1 | A, L)
    """
    # Trying to fit Machine Learning model
    try:
        fm = ml_model.fit(X=xdata, y=ydata)
    except TypeError:
        raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                        "covers both sklearn and supylearner. If there is a predictive model you would "
                        "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                        "can work on adding support")
    if print_results and hasattr(fm, 'summary'):  # Nice summarize option from SuPyLearner
        print('==============================================================================')
        print('Outcome Model')
        fm.summary()
        print('==============================================================================')

    # Generating predictions
    if continuous:
        if hasattr(fm, 'predict'):
            qa = fm.predict(xdata)
            return qa, fm
        else:
            raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")

    else:
        if hasattr(fm, 'predict_proba'):
            qa = fm.predict_proba(xdata)
            if qa.ndim == 1:  # Allows for PyGAM
                return qa, fm
            else:
                return qa[:, 1], fm
        elif hasattr(fm, 'predict'):
            qa = fm.predict(xdata)
            return qa, fm
        else:
            raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")


def stochastic_outcome_predict(xdata, fit_ml_model, continuous):
    """Function to generate predictions machine learning predictions. Used by StochasticTMLE to generate predicted
    probabilities of outcome (i.e. Pr(Y=1 | A=a*, L) in the Monte-Carlo integration procedure
    """
    # Generating predictions
    if continuous:
        if hasattr(fit_ml_model, 'predict'):
            qa_star = fit_ml_model.predict(xdata)
            return qa_star
        else:
            raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")
    else:
        if hasattr(fit_ml_model, 'predict_proba'):
            qa_star = fit_ml_model.predict_proba(xdata)
            if qa_star.ndim == 1:  # Allows for PyGAM
                return qa_star
            else:
                return qa_star[:, 1]
        elif hasattr(fit_ml_model, 'predict'):
            qa_star = fit_ml_model.predict(xdata)
            return qa_star
        else:
            raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")


def missing_machine_learner(xdata, mdata, all_a, none_a, ml_model, print_results=True):
    """Function to fit machine learning predictions. Used by TMLE to generate predicted probabilities of missing
     outcome data, Pr(M=1|A,L)
    """
    # Trying to fit the Machine Learning model
    try:
        fm = ml_model.fit(X=xdata, y=mdata)
    except TypeError:
        raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                        "covers both sklearn and supylearner. If there is a predictive model you would "
                        "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                        "can work on adding support")
    if print_results and hasattr(fm, 'summary'):  # SuPyLearner has a nice summarize function
        print('==============================================================================')
        print('Censoring Model')
        fm.summary()
        print('==============================================================================')

    # Generating predictions
    if hasattr(fm, 'predict_proba'):
        ma1 = fm.predict_proba(all_a)
        ma0 = fm.predict_proba(none_a)
        if (ma1.ndim == 1) and (ma0.ndim == 1):
            return ma1, ma0
        else:
            return ma1[:, 1], ma0[:, 1]
    elif hasattr(fm, 'predict'):
        ma1 = fm.predict(all_a)
        ma0 = fm.predict(none_a)
        return ma1, ma0
    else:
        raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")


def iptw_calculator(df, treatment, model_denom, model_numer, weight, stabilized, standardize, bound, print_results):
    """Background function to calculate inverse probability of treatment weights. Used by `IPTW`, `AIPTW`, `IPSW`,
    `AIPSW`
    """
    denominator_model = propensity_score(df, treatment + ' ~ ' + model_denom,
                                         weights=weight, print_results=print_results)
    d = denominator_model.predict(df)

    # Calculating numerator probabilities (if stabilized)
    if stabilized is True:
        numerator_model = propensity_score(df, treatment + ' ~ ' + model_numer,
                                           weights=weight, print_results=print_results)
        n = numerator_model.predict(df)
    else:
        if model_numer != '1':
            raise ValueError('Argument for model_numerator is only used for stabilized=True')
        n = 1

    # Bounding predicted probabilities if requested
    if bound:
        d = probability_bounds(d, bounds=bound)
        n = probability_bounds(n, bounds=bound)

    # Calculating weights
    if stabilized:  # Stabilized weights
        if standardize == 'population':
            iptw = np.where(df[treatment] == 1, (n / d), ((1 - n) / (1 - d)))
            iptw = np.where(df[treatment].isna(), np.nan, iptw)
        # Stabilizing to exposed (compares all exposed if they were exposed versus unexposed)
        elif standardize == 'exposed':
            iptw = np.where(df[treatment] == 1, 1, (d / (1 - d)) * ((1 - n) / n))
            iptw = np.where(df[treatment].isna(), np.nan, iptw)
        # Stabilizing to unexposed (compares all unexposed if they were exposed versus unexposed)
        else:
            iptw = np.where(df[treatment] == 1, (((1 - d) / d) * (n / (1 - n))), 1)
            iptw = np.where(df[treatment].isna(), np.nan, iptw)

    else:  # Unstabilized weights
        if standardize == 'population':
            iptw = np.where(df[treatment] == 1, 1 / d, 1 / (1 - d))
            iptw = np.where(df[treatment].isna(), np.nan, iptw)
        # Stabilizing to exposed (compares all exposed if they were exposed versus unexposed)
        elif standardize == 'exposed':
            iptw = np.where(df[treatment] == 1, 1, (d / (1 - d)))
            iptw = np.where(df[treatment].isna(), np.nan, iptw)
        # Stabilizing to unexposed (compares all unexposed if they were exposed versus unexposed)
        else:
            iptw = np.where(df[treatment] == 1, ((1 - d) / d), 1)
            iptw = np.where(df[treatment].isna(), np.nan, iptw)
    return d, n, iptw


def plot_kde(df, treatment, probability,
             measure='probability', bw_method='scott', fill=True, color_e='b', color_u='r'):
    """Generates a density plot that can be used to check whether positivity may be violated qualitatively. The
    kernel density used is SciPy's Gaussian kernel. Either Scott's Rule or Silverman's Rule can be implemented.
    Alternative option to the boxplot of probabilities

    Parameters
    ------------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    treatment : str
        Column name of the treatment variable
    probability : str
        Column name of the predicted probability of treatment
    measure : str, optional
        Measure to plot. Options include either the probabilities or log-odds stratified by treatment received.
        Default is probabilities (measure='probability'). Log-odds can be requested via measure='logit'
    bw_method : str, optional
        Method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
    fill : bool, optional
        Whether to color the area under the density curves. Default is true
    color_e : str, optional
        Color of the line/area for the treated group. Default is Blue
    color_u : str, optional
        Color of the line/area for the treated group. Default is Red

    Returns
    ---------------
    matplotlib axes
    """
    if measure == 'probability':
        x = np.linspace(0, 1, 10000)
        density_t = gaussian_kde(df.loc[df[treatment] == 1][probability].dropna(),
                                 bw_method=bw_method)
        density_u = gaussian_kde(df.loc[df[treatment] == 0][probability].dropna(),
                                 bw_method=bw_method)
    elif measure == 'logit':
        t = np.log(probability_to_odds(df.loc[df[treatment] == 1][probability].dropna()))
        density_t = gaussian_kde(t, bw_method=bw_method)

        u = np.log(probability_to_odds(df.loc[df[treatment] == 0][probability].dropna()))
        density_u = gaussian_kde(u, bw_method=bw_method)
        x = np.linspace(np.min((np.min(t), np.min(u))) - 1, np.max((np.max(t), np.max(u))) + 1, 10000)
    else:
        raise ValueError("Only plots of probabilities or log-odds are supported. Please specify either "
                         "'probability' or 'logit'")

    ax = plt.gca()
    if fill:
        ax.fill_between(x, density_t(x), color=color_e, alpha=0.2, label=None)
        ax.fill_between(x, density_u(x), color=color_u, alpha=0.2, label=None)
    ax.plot(x, density_t(x), color=color_e, label='Treat = 1')
    ax.plot(x, density_u(x), color=color_u, label='Treat = 0')
    if measure == 'probability':
        ax.set_xlabel('Probability')
    else:
        ax.set_xlabel('Log-Odds')
    ax.set_ylabel('Density')
    ax.legend()
    return ax


def plot_boxplot(df, treatment, probability, measure='probability'):
    """Generates a stratified boxplot that can be used to visually check whether positivity may be violated,
    qualitatively. Alternative option to the kernel density plot.

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    treatment : str
        Column name of the treatment variable
    probability : str
        Column name of the predicted probability of treatment

    measure : str, optional
        Measure to plot. Options include either the probabilities or log-odds stratified by treatment received.
        Default is probabilities (measure='probability'). Log-odds can be requested via measure='logit'

    Returns
    -------------
    matplotlib axes
    """
    if measure == 'probability':
        boxes = (df.loc[df[treatment] == 1][probability].dropna(),
                 df.loc[df[treatment] == 0][probability].dropna())

    elif measure == 'logit':
        boxes = (np.log(probability_to_odds(df.loc[df[treatment] == 1][probability].dropna())),
                 np.log(probability_to_odds(df.loc[df[treatment] == 0][probability].dropna())))
    else:
        raise ValueError("Only plots of probabilities or log-odds are supported. Please specify either "
                         "'probability' or 'logit")

    labs = ['A = 1', 'A = 0']
    meanpointprops = dict(marker='D', markeredgecolor='black', markerfacecolor='black')
    ax = plt.gca()
    ax.boxplot(boxes, labels=labs, meanprops=meanpointprops, showmeans=True)
    if measure == 'probability':
        ax.set_ylabel('Probability')
        ax.set_ylim([0, 1])
    else:
        ax.set_ylabel('Log-Odds')
    return ax


def positivity(df, weights):
    """Use this to assess whether positivity is a valid assumption. For stabilized weights, the mean weight should
    be approximately 1. For unstabilized weights, the mean weight should be approximately 2. If there are extreme
    outliers, this may indicate problems with the calculated weights

    Parameters
    --------------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    weights : str
        Column name of the inverse probability of treatment weights

    Returns
    --------------
    tuple
        Tuple of positivity results; mean, SD, min, max
    """
    pos_avg = float(np.mean(df[weights].dropna()))
    pos_max = np.max(df[weights].dropna())
    pos_min = np.min(df[weights].dropna())
    pos_sd = float(np.std(df[weights].dropna()))
    return pos_avg, pos_sd, pos_min, pos_max


def standardized_mean_differences(df, treatment, weight, formula):
    """Calculates the standardized mean differences for all variables. Default calculates the standardized mean
    difference for all variables included in the IPTW denominator

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    treatment : str
        Column name of the treatment variable
    weight : str
        Column name of the inverse probability of treatment weights
    formula : str
        Formula for weights model following patsy syntax

    Returns
    -------
    DataFrame
        Returns pandas DataFrame of calculated standardized mean differences. Columns are labels (variables labels),
        smd_u (unweighted standardized difference), and smd_w (weighted standardized difference)
    iptw_only : bool, optional
        Whether the diagnostic should be run on IPTW only or the weights multiplied together. Default is IPTW only
    """
    def _standardized_difference_(df, treatment, var_type, weight, weighted=True):
        """Background function to calculate the standardized mean difference between the treat and untreated for a
        specified variable. Useful for checking whether a confounder was balanced between the two treatment groups
        by the specified IPTW model SMD based on: Austin PC 2011; https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3144483/
        """
        def _categorical_cov_(a, b):
            """Turns out, pandas and numpy don't have the correct covariance matrix I need for categorical variables.
            The covariance matrix is defined as

            S = [S_{kl}] = (P_{1k}*(1-P_{1k}) + P_{2k}*(1-P{2k})) / 2     if k == l
                           (P_{1k}*P_{1l} + P_{2k}*P_{2l}) / 2            if k != l
            """
            cv2 = []
            for i, v in enumerate(a):
                cv1 = []
                if i == 0:
                    pass
                else:
                    for j, w in enumerate(b):
                        if j == 0:
                            pass
                        elif i == j:
                            cv1.append((v * (1 - v) + w * (1 - w)) / 2)
                        else:
                            cv1.append((a[i] * a[j] + b[i] * b[j]) / -2)
                    cv2.append(cv1)

            return np.array(cv2)

        # Pulling out relevant data
        dft = df.loc[(df[treatment] == 1) & (df[weight].notnull())].copy()
        dfn = df.loc[(df[treatment] == 0) & (df[weight].notnull())].copy()
        vcols = list(df.columns)
        vcols.remove(treatment)
        vcols.remove(weight)

        if var_type == 'binary':
            if weighted:
                dwt = DescrStatsW(dft[vcols], weights=dft[weight])
                wt = dwt.mean
                dwn = DescrStatsW(dfn[vcols], weights=dfn[weight])
                wn = dwn.mean
            else:
                wt = np.mean(dft[vcols].dropna(), axis=0)
                wn = np.mean(dfn[vcols].dropna(), axis=0)
            return float((wt - wn) / np.sqrt((wt * (1 - wt) + wn * (1 - wn)) / 2))

        elif var_type == 'continuous':
            if weighted:
                dwt = DescrStatsW(dft[vcols], weights=dft[weight], ddof=1)
                wmt = dwt.mean
                wst = dwt.std
                dwn = DescrStatsW(dfn[vcols], weights=dfn[weight], ddof=1)
                wmn = dwn.mean
                wsn = dwn.std
            else:
                dwt = DescrStatsW(dft[vcols], ddof=1)
                wmt = dwt.mean
                wst = dwt.std
                dwn = DescrStatsW(dfn[vcols], ddof=1)
                wmn = dwn.mean
                wsn = dwn.std
            return float((wmt - wmn) / np.sqrt((wst ** 2 + wsn ** 2) / 2))

        elif var_type == 'categorical':
            if weighted:
                wt = np.average(dft[vcols], weights=dft[weight], axis=0)
                wn = np.average(dfn[vcols], weights=dfn[weight], axis=0)
            else:
                wt = np.mean(dft[vcols], axis=0)
                wn = np.mean(dfn[vcols], axis=0)

            t_c = wt - wn
            s_inv = np.linalg.inv(_categorical_cov_(a=wt, b=wn))
            return float(np.sqrt(np.dot(np.transpose(t_c[1:]), np.dot(s_inv, t_c[1:]))))

        else:
            raise ValueError('Not supported')

    variables = patsy.dmatrix(formula + ' - 1', df, return_type='dataframe')
    w_diff = []
    u_diff = []
    vlabel = []

    # Pull out list of terms and the corresponding dataframe slice(s)
    term_dict = variables.design_info.term_name_slices

    # Looping through the terms
    for term in variables.design_info.terms:
        # Adding term labels
        vlabel.append(term.name())

        # Pulling out data corresponding to term
        chunk = term_dict[term.name()]
        v = variables.iloc[:, chunk].copy()

        # Detecting variable type
        if v.shape[1] != 1:
            vtype = 'categorical'
        elif np.all(v.dropna().isin([0, 1])):
            vtype = 'binary'
        else:
            vtype = 'continuous'

        # calculate the absolute standardized difference
        dat = pd.concat([v, df[[treatment, weight]]], axis=1)
        wsmd = _standardized_difference_(df=dat, treatment=treatment, var_type=vtype, weight=weight, weighted=True)
        w_diff.append(wsmd)
        usmd = _standardized_difference_(df=dat, treatment=treatment, var_type=vtype, weight=weight, weighted=False)
        u_diff.append(usmd)

    # Setting up DataFrame to return with calculated differences
    s = pd.DataFrame()
    s['labels'] = vlabel
    s['smd_w'] = w_diff
    s['smd_u'] = u_diff
    return s


def plot_love(df, treatment, weight, formula,
              color_unweighted='r', color_weighted='b', shape_unweighted='o', shape_weighted='o'):
    """Generates a Love-plot to detail covariate balance based on the IPTW weights. Further details on the usage of
    this plot are available in Austin PC & Stuart EA 2015 https://onlinelibrary.wiley.com/doi/full/10.1002/sim.6607

    The Love plot generates a dashed line at standardized mean difference of 0.10. Ideally, weighted SMD are below
    this level. Below 0.20 may also be sufficient. Variables above this level may be unbalanced despite the
    weighting procedure. Different functional forms (or approaches like machine learning) may be worth considering

    Parameters
    ----------
    df : dataframe
        Pandas DataFrame object containing variables of interest
    treatment : str
        Treatment/exposure variable column label
    weight : str
        Column label for the weights
    formula : str
        Right-hand side of the equation for the weights
    color_unweighted : str, optional
        Color for the unweighted standardized mean differences. Default is red
    color_weighted : str, optional
        Color for the weighted standardized mean differences. Default is blue
    shape_unweighted : str, optional
        Shape of points for the unweighted standardized mean differences. Default is circles
    shape_weighted:
        Shape of points for the weighted standardized mean differences. Default is circles

    Returns
    -------
    axes
        Matplotlib axes of the Love plot
    """
    to_plot = standardized_mean_differences(df=df, treatment=treatment, weight=weight, formula=formula)
    to_plot['smd_w'] = np.absolute(to_plot['smd_w'])
    to_plot['smd_u'] = np.absolute(to_plot['smd_u'])
    to_plot = to_plot.sort_values(by='smd_u', ascending=True).reset_index(drop=True)

    # Generate plot
    ax = plt.gca()
    ax.plot(to_plot.smd_u, to_plot.index, shape_unweighted, c=color_unweighted, label='Unweighted')
    ax.plot(to_plot.smd_w, to_plot.index, shape_weighted, c=color_weighted, label='Weighted')
    ax.set_xlim([0, np.max([np.max(to_plot['smd_w']), np.max(to_plot['smd_u'])]) + 0.5])
    ax.set_xlabel('Absolute Standardized Difference')
    ax.axvline(0.1, color='gray')
    ax.set_yticks([i for i in range(to_plot.shape[0])])
    ax.set_yticklabels(to_plot['labels'])
    ax.legend()
    return ax


def outcome_accuracy(true, predicted, decimal=3):
    """Diagnostic for Q-models. Compares the observed outcome to the predicted.

    Parameters
    ----------
    true:
        True (observed) values
    predicted:
        Predicted values
    decimal : int, optional
        Number of decimal places to display. Default is three
    """
    value = predicted - true

    print('======================================================================')
    print('                 Natural Course Prediction Accuracy')
    print('======================================================================')
    print('Outcome model accuracy summary statistics. Defined as the predicted\n'
          'outcome value minus the observed outcome value')
    print('----------------------------------------------------------------------')
    print('Mean value:           ', np.round(np.mean(value), decimal))
    print('Standard Deviation:    ', np.round(np.std(value, ddof=1), decimal))
    print('Minimum value:        ', np.round(np.min(value), decimal))
    print('Maximum value:        ', np.round(np.max(value), decimal))
    print('======================================================================')


def plot_kde_accuracy(values, bw_method='scott', fill=True, color='b'):
    """Generates a density plot that can be used to check whether positivity may be violated qualitatively. The
    kernel density used is SciPy's Gaussian kernel. Either Scott's Rule or Silverman's Rule can be implemented.
    Alternative option to the boxplot of probabilities

    Parameters
    ----------
    values : str, optional
        Difference between predicted value and observed value
    bw_method : str, optional
        Method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
    fill : bool, optional
        Whether to color the area under the density curves. Default is true
    color : str, optional
        Color of the line/area. Default is blue

    Returns
    -------
    matplotlib axes
    """
    x = np.linspace(np.min(values) - 0.5, np.max(values) + 0.5, 10000)
    density = gaussian_kde(values, bw_method=bw_method)

    ax = plt.gca()
    if fill:
        ax.fill_between(x, density(x), color=color, alpha=0.2, label=None)
    ax.plot(x, density(x), color=color)
    ax.set_xlabel(r'$Y_{pred} - Y$')
    ax.set_ylabel('Density')
    return ax


def stochastic_check_conditional(df, conditional):
    """Check that conditionals are exclusive for the stochastic fit process. Generates a warning if not true
    """
    a = np.zeros(df.shape[0])
    for c in conditional:
        a = a + np.where(eval(c), 1, 0)

    if np.any(a > 1):
        warnings.warn("It looks like your conditional categories are NOT exclusive. For appropriate estimation, "
                      "the conditions that designate each category should be exclusive", UserWarning)


def aipw_calculator(y, a, py_a, py_n, pa1, pa0, difference=True, weights=None, splits=None, continuous=False):
    """Function to calculate AIPW estimates. Called by AIPTW, SingleCrossfitAIPTW, and DoubleCrossfitAIPTW
    """
    # Point estimate calculation
    y1 = np.where(a == 1, (y - py_a*(1 - pa1)) / pa1, py_a)
    y0 = np.where(a == 0, (y - py_n*(1 - pa0)) / pa0, py_n)

    # Warning system if values are out of range
    if not continuous:
        if np.mean(y1) > 1 or np.mean(y1) < 0:
            warnings.warn("The estimated probability for all-exposed is out of the bounds (less than zero or greater "
                          "than 1). This may indicate positivity issues resulting from extreme weights, too small of a "
                          "sample size, or too flexible of models. Try setting the optional `bound` argument. If using "
                          "DoubleCrossfitAIPTW, try SingleCrossfitAIPTW or the TMLE estimators instead.", UserWarning)
        if np.mean(y0) > 1 or np.mean(y0) < 0:
            warnings.warn("The estimated probability for none-exposed is out of the bounds (less than zero or greater "
                          "than 1). This may indicate positivity issues resulting from extreme weights, too small of a "
                          "sample size, or too flexible of models. Try setting the optional `bound` argument. If using "
                          "DoubleCrossfitAIPTW, try SingleCrossfitAIPTW or the TMLE estimators instead.", UserWarning)

    # Calculating ACE as a difference
    if difference:
        if weights is None:
            estimate = np.nanmean(y1 - y0)
            if splits is None:
                var = np.nanvar((y1 - y0) - estimate, ddof=1) / y.shape[0]
            else:
                var_rd = []
                for i in set(splits):
                    y1s = y1[splits == i]
                    y0s = y0[splits == i]
                    var_rd.append(np.var((y1s - y0s) - estimate, ddof=1))
                var = np.mean(var_rd) / y.shape[0]
        else:
            estimate = DescrStatsW(y1, weights=weights).mean - DescrStatsW(y0, weights=weights).mean
            var = np.nan

    # Calculating ACE as a ratio
    else:
        if weights is None:
            estimate = np.nanmean(y1) / np.nanmean(y0)
            if estimate < 0:
                warnings.warn("lower than 0", UserWarning)
            py_o = a*py_a + (1-a)*py_n
            ic = ((a*(y-py_o)) / (np.mean(py_a)*pa1) + (py_a - np.mean(py_a)) -
                  ((1-a)*(y-py_o)) / (np.mean(py_n)*pa0) + (py_n - np.mean(py_n)))
            var = np.nanvar(ic, ddof=1) / y.shape[0]
        else:
            estimate = DescrStatsW(y1, weights=weights).mean / DescrStatsW(y0, weights=weights).mean
            var = np.nan

    return estimate, var
